"""
Routines for reading ODIM_H5 files.

"""

import datetime

import numpy as np
try:
    import h5py
    _H5PY_AVAILABLE = True
except ImportError:
    _H5PY_AVAILABLE = False

from ..config import FileMetadata, get_fillvalue
from ..io.common import make_time_unit_str, _test_arguments
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency


ODIM_H5_FIELD_NAMES = {
    'TH': 'total_power_horizontal',        # uncorrected reflectivity, horizontal
    'TV': 'total_power_vertical',        # uncorrected reflectivity, vertical
    'DBZH': 'reflectivity_horizontal',     # corrected reflectivity, horizontal
    'DBZV': 'reflectivity_vertical',     # corrected reflectivity, vertical
    'ZDR': 'differential_reflectivity',     # differential reflectivity
    'RHOHV': 'cross_correlation_ratio',
    'LDR': 'linear_polarization_ratio',
    'PHIDP': 'differential_phase',
    'KDP': 'specific_differential_phase',
    'SQI': 'normalized_coherent_power',
    'SNR': 'signal_to_noise_ratio',
    'VRAD': 'velocity',  # radial velocity, marked for deprecation in ODIM HDF5 2.2
    'VRADH': 'velocity_horizontal',  # radial velocity, horizontal polarisation
    'VRADV': 'velocity_vertical',  # radial velocity, vertical polarisation
    'WRAD': 'spectrum_width',
    'QIND': 'quality_index',
}


def read_odim_h5(filename, field_names=None, additional_metadata=None,
                 file_field_names=False, exclude_fields=None,
                 include_fields=None, include_datasets=None,
                 exclude_datasets=None, **kwargs):
    """
    Read a ODIM_H5 file.

    Parameters
    ----------
    filename : str
        Name of the ODIM_H5 file to read.
    field_names : dict, optional
        Dictionary mapping ODIM_H5 field names to radar field names. If a
        data type found in the file does not appear in this dictionary or has
        a value of None it will not be placed in the radar.fields dictionary.
        A value of None, the default, will use the mapping defined in the
        Py-ART configuration file.
    additional_metadata : dict of dicts, optional
        Dictionary of dictionaries to retrieve metadata from during this read.
        This metadata is not used during any successive file reads unless
        explicitly included.  A value of None, the default, will not
        introduct any addition metadata and the file specific or default
        metadata as specified by the Py-ART configuration file will be used.
    file_field_names : bool, optional
        True to use the MDV data type names for the field names. If this
        case the field_names parameter is ignored. The field dictionary will
        likely only have a 'data' key, unless the fields are defined in
        `additional_metadata`.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    include_datasets : list or None, optional
        List of datasets to include from the HDF5 file, given
        as ["dataset1", "dataset2", ...]. Set to None to include all datasets
        not specified by exclude_datasets.
    exclude_datasets : list or None, optional
        List of datasets to exclude from the HDF5 file, given
        as ["dataset1", "dataset2", ...]. Set to None to include all datasets
        specified by include_datasets.

    Returns
    -------
    radar : Radar
        Radar object containing data from ODIM_H5 file.

    """
    # TODO before moving to pyart.io
    # * unit test
    # * add default field mapping, etc to default config
    # * auto-detect file type with pyart.io.read function
    # * instrument parameters
    # * add additional checks for HOW attributes
    # * support for other objects (SCAN, XSEC)

    # check that h5py is available
    if not _H5PY_AVAILABLE:
        raise MissingOptionalDependency(
            "h5py is required to use read_odim_h5 but is not installed")

    # test for non empty kwargs
    _test_arguments(kwargs)

    # create metadata retrieval object
    if field_names is None:
        field_names = ODIM_H5_FIELD_NAMES
    filemetadata = FileMetadata('odim_h5', field_names, additional_metadata,
                                file_field_names, exclude_fields,
                                include_fields)

    # open the file
    with h5py.File(filename, 'r') as hfile:
        odim_object = _to_str(hfile['what'].attrs['object'])
        if odim_object not in ['PVOL', 'SCAN', 'ELEV', 'AZIM']:
            raise NotImplementedError(
                'object: %s not implemented.' % (odim_object))

        # determine the number of sweeps by the number of groups which
        # begin with dataset
        if include_datasets is not None:
            datasets = [k for k in hfile if k.startswith('dataset')
                        and k in include_datasets]
        elif exclude_datasets is not None:
            datasets = [k for k in hfile if k.startswith('dataset')
                        and k not in exclude_datasets]
        else:
            datasets = [k for k in hfile if k.startswith('dataset')]
        datasets.sort(key=lambda x: int(x[7:]))
        nsweeps = len(datasets)

        # latitude, longitude and altitude
        latitude = filemetadata('latitude')
        longitude = filemetadata('longitude')
        altitude = filemetadata('altitude')

        h_where = hfile['where'].attrs
        latitude['data'] = np.array([h_where['lat']], dtype='float64')
        longitude['data'] = np.array([h_where['lon']], dtype='float64')
        altitude['data'] = np.array([h_where['height']], dtype='float64')

        # metadata
        metadata = filemetadata('metadata')
        metadata['source'] = _to_str(hfile['what'].attrs['source'])
        metadata['original_container'] = 'odim_h5'
        metadata['odim_conventions'] = _to_str(hfile.attrs['Conventions'])

        h_what = hfile['what'].attrs
        metadata['version'] = _to_str(h_what['version'])
        metadata['source'] = _to_str(h_what['source'])

        try:
            ds1_how = hfile[datasets[0]]['how'].attrs
        except KeyError:
            # if no how group exists mock it with an empty dictionary
            ds1_how = {}
        if 'system' in ds1_how:
            metadata['system'] = ds1_how['system']
        if 'software' in ds1_how:
            metadata['software'] = ds1_how['software']
        if 'sw_version' in ds1_how:
            metadata['sw_version'] = ds1_how['sw_version']

        # sweep_start_ray_index, sweep_end_ray_index
        sweep_start_ray_index = filemetadata('sweep_start_ray_index')
        sweep_end_ray_index = filemetadata('sweep_end_ray_index')

        if odim_object in ['AZIM', 'SCAN', 'PVOL']:
            rays_per_sweep = [
                int(hfile[d]['where'].attrs['nrays']) for d in datasets]
        elif odim_object == 'ELEV':
            rays_per_sweep = [
                int(hfile[d]['where'].attrs['angles'].size) for d in datasets]
        total_rays = sum(rays_per_sweep)
        ssri = np.cumsum(np.append([0], rays_per_sweep[:-1])).astype('int32')
        seri = np.cumsum(rays_per_sweep).astype('int32') - 1
        sweep_start_ray_index['data'] = ssri
        sweep_end_ray_index['data'] = seri

        # sweep_number
        sweep_number = filemetadata('sweep_number')
        sweep_number['data'] = np.arange(nsweeps, dtype='int32')

        # sweep_mode
        sweep_mode = filemetadata('sweep_mode')
        sweep_mode['data'] = np.array(nsweeps * ['azimuth_surveillance'])

        # scan_type
        if odim_object == 'ELEV':
            scan_type = 'rhi'
        else:
            scan_type = 'ppi'

        # fixed_angle
        fixed_angle = filemetadata('fixed_angle')
        if odim_object == 'ELEV':
            sweep_el = [hfile[d]['where'].attrs['az_angle'] for d in datasets]
        else:
            sweep_el = [hfile[d]['where'].attrs['elangle'] for d in datasets]
        fixed_angle['data'] = np.array(sweep_el, dtype='float32')

        # elevation
        elevation = filemetadata('elevation')
        if 'elangles' in ds1_how:
            edata = np.empty(total_rays, dtype='float32')
            for d, start, stop in zip(datasets, ssri, seri):
                edata[start:stop+1] = hfile[d]['how'].attrs['elangles'][:]
            elevation['data'] = edata
        elif odim_object == 'ELEV':
            edata = np.empty(total_rays, dtype='float32')
            for d, start, stop in zip(datasets, ssri, seri):
                edata[start:stop+1] = hfile[d]['where'].attrs['angles'][:]
            elevation['data'] = edata
        else:
            elevation['data'] = np.repeat(sweep_el, rays_per_sweep)

        # range
        _range = filemetadata('range')
        if 'rstart' in hfile['dataset1/where'].attrs:
            # derive range from rstart and rscale attributes if available

            # check that the gate spacing is constant between sweeps
            rstart = [hfile[d]['where'].attrs['rstart'] for d in datasets]
            if any(rstart != rstart[0]):
                raise ValueError('range start changes between sweeps')
            rscale = [hfile[d]['where'].attrs['rscale'] for d in datasets]
            if any(rscale != rscale[0]):
                raise ValueError('range scale changes between sweeps')
            all_sweeps_nbins = [hfile[d]['where'].attrs['nbins'] for d in datasets]
            # check for max range off all sweeps
            max_nbins = max(all_sweeps_nbins)

            if isinstance(max_nbins, np.ndarray):
                max_nbins = max_nbins[0]
            else:
                max_nbins = max(all_sweeps_nbins)

            rscenter = 1e3 * rstart[0] + rscale[0] / 2
            _range['data'] = np.arange(rscenter,
                                       rscenter + max_nbins * rscale[0],
                                       rscale[0], dtype='float32')
            _range['meters_to_center_of_first_gate'] = rstart[0] * 1000.
            _range['meters_between_gates'] = float(rscale[0])
        else:
            # if not defined use range attribute which defines the maximum range
            # in km. There is no information on the starting location of the
            # range bins so we assume this to be 0.
            # This most often occurs in RHI files, which technically do not meet
            # the ODIM 2.2 specs. Section 7.4 requires that these files include
            # the where/rstart, where/rscale and where/nbins attributes.
            max_range = [hfile[d]['where'].attrs['range'] for d in datasets]
            if any(max_range != max_range[0]):
                raise ValueError('maximum range changes between sweeps')
            # nbins is required
            nbins = hfile['dataset1/data1/data'].shape[1]
            _range['data'] = np.linspace(
                0, max_range[0] * 1000., nbins).astype('float32')
            _range['meters_to_center_of_first_gate'] = 0
            _range['meters_between_gates'] = max_range[0] * 1000. / nbins

        # azimuth
        azimuth = filemetadata('azimuth')
        az_data = np.ones((total_rays, ), dtype='float32')
        for dset, start, stop in zip(datasets, ssri, seri):
            if odim_object == 'ELEV':
                # all azimuth angles are the sweep azimuth angle
                sweep_az = hfile[dset]['where'].attrs['az_angle']
            elif odim_object == 'AZIM':
                # Sector azimuths are specified in the startaz and stopaz
                # attribute of dataset/where.
                # Assume that the azimuth angles do not pass through 0/360 deg.
                startaz = hfile[dset]['where'].attrs['startaz']
                stopaz = hfile[dset]['where'].attrs['stopaz']
                nrays = stop - start + 1
                sweep_az = np.linspace(startaz, stopaz, nrays, endpoint=True)
            elif ('startazA' in ds1_how) and ('stopazA' in ds1_how):
                # average between start and stop azimuth angles
                startaz = hfile[dset]['how'].attrs['startazA']
                stopaz = hfile[dset]['how'].attrs['stopazA']
                sweep_az = np.angle(
                    (np.exp(1.j*np.deg2rad(startaz)) +
                    np.exp(1.j*np.deg2rad(stopaz))) / 2., deg=True)
            else:
                # according to section 5.1 the first ray points to the
                # northernmost direction and proceeds clockwise for a complete
                # 360 rotation.
                try:
                    astart = hfile[dset]['how'].attrs['astart']
                except KeyError:
                    astart = 0.0
                nrays = hfile[dset]['where'].attrs['nrays']
                da = 360.0 / nrays
                sweep_az = np.arange(astart + da / 2., 360., da, dtype='float32')
            az_data[start:stop+1] = sweep_az
        azimuth['data'] = az_data

        # time
        _time = filemetadata('time')
        if ('startazT' in ds1_how) and ('stopazT' in ds1_how):
            # average between startazT and stopazT
            t_data = np.empty((total_rays, ), dtype='float32')
            for dset, start, stop in zip(datasets, ssri, seri):
                t_start = hfile[dset]['how'].attrs['startazT']
                t_stop = hfile[dset]['how'].attrs['stopazT']
                t_data[start:stop+1] = (t_start + t_stop) / 2
            start_epoch = t_data.min()
            start_time = datetime.datetime.utcfromtimestamp(start_epoch)
            _time['units'] = make_time_unit_str(start_time)
            _time['data'] = t_data - start_epoch
        else:
            t_data = np.empty((total_rays, ), dtype='int32')
            # interpolate between each sweep starting and ending time
            for dset, start, stop in zip(datasets, ssri, seri):
                dset_what = hfile[dset]['what'].attrs
                start_str = _to_str(
                    dset_what['startdate'] + dset_what['starttime'])
                end_str = _to_str(dset_what['enddate'] + dset_what['endtime'])
                start_dt = datetime.datetime.strptime(start_str, '%Y%m%d%H%M%S')
                end_dt = datetime.datetime.strptime(end_str, '%Y%m%d%H%M%S')

                time_delta = end_dt - start_dt
                delta_seconds = time_delta.seconds + time_delta.days * 3600 * 24
                rays = stop - start + 1
                sweep_start_epoch = (
                    start_dt - datetime.datetime(1970, 1, 1)).total_seconds()
                t_data[start:stop+1] = (sweep_start_epoch +
                                        np.linspace(0, delta_seconds, rays))
            start_epoch = t_data.min()
            start_time = datetime.datetime.utcfromtimestamp(start_epoch)
            _time['units'] = make_time_unit_str(start_time)
            _time['data'] = (t_data - start_epoch).astype('float32')

        # fields
        fields = {}
        h_field_keys = [k for k in hfile['dataset1'] if k.startswith('data')]
        odim_fields = [hfile['dataset1'][d]['what'].attrs['quantity'] for d in
                    h_field_keys]
        for odim_field, h_field_key in zip(odim_fields, h_field_keys):
            field_name = filemetadata.get_field_name(_to_str(odim_field))
            if field_name is None:
                continue
            fdata = np.ma.zeros((total_rays, max_nbins), dtype='float32')
            start = 0
            # loop on the sweeps, copy data into correct location in data array
            for dset, rays_in_sweep in zip(datasets, rays_per_sweep):
                try:
                    sweep_data = _get_odim_h5_sweep_data(hfile[dset][h_field_key])
                except KeyError:
                    sweep_data = np.zeros((rays_in_sweep, max_nbins)) + np.NaN
                sweep_nbins = sweep_data.shape[1]
                fdata[start:start + rays_in_sweep, :sweep_nbins] = sweep_data[:]
                # set data to NaN if its beyond the range of this sweep
                fdata[start:start + rays_in_sweep,
                      sweep_nbins:max_nbins] = np.nan
                start += rays_in_sweep
            # create field dictionary
            field_dic = filemetadata(field_name)
            field_dic['data'] = fdata
            field_dic['_FillValue'] = get_fillvalue()
            fields[field_name] = field_dic

    # instrument_parameters
    instrument_parameters = None

    return Radar(
        _time, _range, fields, metadata, scan_type,
        latitude, longitude, altitude,
        sweep_number, sweep_mode, fixed_angle, sweep_start_ray_index,
        sweep_end_ray_index,
        azimuth, elevation,
        instrument_parameters=instrument_parameters)


def _to_str(text):
    """ Convert bytes to str if necessary. """
    if hasattr(text, 'decode'):
        return text.decode('utf-8')
    else:
        return text


def _get_odim_h5_sweep_data(group):
    """ Get ODIM_H5 sweet data from an HDF5 group. """

    # mask raw data
    what = group['what']
    raw_data = group['data'][:]

    if 'nodata' in what.attrs:
        nodata = what.attrs.get('nodata')
        data = np.ma.masked_equal(raw_data, nodata)
    else:
        data = np.ma.masked_array(raw_data)
    if 'undetect' in what.attrs:
        undetect = what.attrs.get('undetect')
        data[data == undetect] = np.ma.masked

    offset = 0.0
    gain = 1.0
    if 'offset' in what.attrs:
        offset = what.attrs.get('offset')
    if 'gain' in what.attrs:
        gain = what.attrs.get('gain')
    return data * gain + offset



""" Unit Tests for Py-ART's retrieve/vad.py module. """

import numpy as np
from numpy.testing import assert_allclose

import pyart


def test_vad_michelson():
    test_radar = pyart.testing.make_target_radar()
    height = np.arange(0.0, 1000.0, 200.0)
    speed = np.ones_like(height) * 5.0
    direction = np.array([0.0, 90.0, 180.0, 270.0, 45.0])
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    sim_vel = pyart.util.simulated_vel_from_profile(test_radar, profile)
    test_radar.add_field('velocity', sim_vel,
                         replace_existing=True)

    velocity = 'velocity'
    z_want = np.linspace(0.0, 10.0, 5)

    vad_height = ([0.0, 2.5, 5.0, 7.5, 10.0])
    vad_speed = ([4.9997, 4.9445, 4.8865,
                  4.8182, 4.7497])
    vad_direction = ([89.8712, 90.5113, 91.2084,
                      92.0622, 92.9569])
    u_wind = ([-4.9997, -4.9443,
               -4.8854, -4.8150, -4.7434])
    v_wind = ([-0.0112, 0.0441,
               0.1030, 0.1733, 0.2450])

    vad = pyart.retrieve.vad_michelson(
        test_radar, velocity, z_want)

    assert_allclose(vad.height, vad_height, rtol=1e-3, atol=1e-1)
    assert_allclose(vad.speed, vad_speed, rtol=1e-3, atol=1e-1)
    assert_allclose(vad.direction, vad_direction,rtol=1e-3, atol=1e-1)
    assert_allclose(vad.u_wind, u_wind, rtol=1e-3, atol=1e-1)
    assert_allclose(vad.v_wind, v_wind, rtol=1e-3, atol=1e-1)


def test_vad_browning():
    test_radar = pyart.testing.make_target_radar()
    height = np.arange(0.0, 1000.0, 200.0)
    speed = np.ones_like(height) * 5.0
    direction = np.array([0.0, 90.0, 180.0, 270.0, 45.0])
    profile = pyart.core.HorizontalWindProfile(height, speed, direction)
    sim_vel = pyart.util.simulated_vel_from_profile(test_radar, profile)
    test_radar.add_field('velocity', sim_vel,
                         replace_existing=True)

    velocity = 'velocity'
    z_want = np.linspace(0.0, 10.0, 5)

    vad_height = ([0.0, 2.5, 5.0, 7.5, 10.0])
    vad_speed = ([4.9728, 4.9465, 4.8802, 4.82951,
                  4.7578])

    vad_direction = ([90.3142, 90.6225, 91.4236, 92.0601,
                      92.99520])

    u_wind = ([-4.9727, -4.9462, -4.8787, -4.8263, -4.7513])
    v_wind = ([0.02727, 0.05374, 0.1212, 0.1736, 0.2486])

    vad = pyart.retrieve.vad_browning(
        test_radar, velocity, z_want)
    assert_allclose(vad.height, vad_height, rtol=1e-3, atol=1e-1)
    assert_allclose(vad.speed, vad_speed, rtol=1e-3, atol=1e-1)
    assert_allclose(vad.direction, vad_direction,rtol=1e-3, atol=1e-1)
    assert_allclose(vad.u_wind, u_wind, rtol=1e-3, atol=1e-1)
    assert_allclose(vad.v_wind, v_wind, rtol=1e-3, atol=1e-1)



#! /usr/bin/env python

import netCDF4

dset = netCDF4.Dataset('sgpinterpolatedsondeC1.c1.20110510.000000.cdf')
dvars = dset.variables

out = netCDF4.Dataset('example_interpolatedsonde.cdf', mode='w')

out.createDimension('time', None)
out.createDimension('height', 316)

time = out.createVariable('time', dvars['time'].dtype, ('time'))
height = out.createVariable('height', dvars['height'].dtype, ('height'))
wspd = out.createVariable('wspd', dvars['wspd'].dtype, ('time', 'height'))
wdir = out.createVariable('wdir', dvars['wdir'].dtype, ('time', 'height'))

time[:] = dvars['time'][685:695]
time.units = dvars['time'].units
height[:] = dvars['height'][:]
wspd[:] = dvars['wspd'][685:695, :]
wdir[:] = dvars['wdir'][685:695, :]

out.close()



"""
GAMICFile class and utility functions.

"""

import numpy as np
import h5py


class GAMICFile(object):
    """
    A class to read GAMIC files.

    Parameters
    ----------
    filename : str
        Filename of GAMIC HDF5 file.

    Attributes
    ----------
    nsweeps : int
        Number of sweeps (or scans) in the file.
    rays_per_sweep : array of int32
        Number of rays in each sweep.
    total_rays : int
        Total number of rays in all sweeps.
    start_ray, end_ray : array of int32
        Index of the first (start) and last (end) ray in each sweep, 0-based.
    _hfile : HDF5 file
        Open HDF5 file object from which data is read.
    _scans : list
        Name of the HDF5 group for each scan.

    """

    def __init__(self, filename):
        """ initialize object. """
        self._hfile = h5py.File(filename, 'r')
        self.nsweeps = self._hfile['what'].attrs['sets']
        self._scans = ['scan%i' % (i) for i in range(self.nsweeps)]
        self.rays_per_sweep = self.how_attrs('ray_count', 'int32')
        self.total_rays = sum(self.rays_per_sweep)
        # starting and ending ray for each sweep
        self.start_ray = np.cumsum(np.append([0], self.rays_per_sweep[:-1]))
        self.end_ray = np.cumsum(self.rays_per_sweep) - 1
        return

    def close(self):
        """ Close the file. """
        self._hfile.close()

    # file checks
    def is_file_complete(self):
        """ True if all scans in file, False otherwise. """
        # check than scan0, scan1, ... scan[nsweeps-1] are present
        for scan in self._scans:
            if scan not in self._hfile:
                return False
        return True

    def is_file_single_scan_type(self):
        """ True is all scans are the same scan type, False otherwise. """
        scan_type = self._hfile['scan0/what'].attrs['scan_type']
        for scan in self._scans:
            if self._hfile[scan]['what'].attrs['scan_type'] != scan_type:
                return False
        return True

    # attribute look up
    def where_attr(self, attr, dtype):
        """ Return an array containing a attribute from the where group. """
        return np.array([self._hfile['where'].attrs[attr]], dtype=dtype)

    def how_attr(self, attr, dtype):
        """ Return an array containing a attribute from the how group. """
        return np.array([self._hfile['how'].attrs[attr]], dtype=dtype)

    def is_attr_in_group(self, group, attr):
        """ True is attribute is present in the group, False otherwise. """
        return attr in self._hfile[group].attrs

    def raw_group_attr(self, group, attr):
        """ Return an attribute from a group with no reformatting. """
        return self._hfile[group].attrs[attr]

    def raw_scan0_group_attr(self, group, attr):
        """ Return an attribute from the scan0 group with no reformatting. """
        return self._hfile['/scan0'][group].attrs[attr]

    # scan/sweep based attribute lookup
    def how_attrs(self, attr, dtype):
        """ Return an array of an attribute for each scan's how group. """
        return np.array([self._hfile[s]['how'].attrs[attr]
                         for s in self._scans], dtype=dtype)

    def how_ext_attrs(self, attr):
        """
        Return a list of an attribute in each scan's how/extended group.
        """
        return [float(self._hfile[s]['how']['extended'].attrs[attr])
                for s in self._scans]

    def what_attrs(self, attr, dtype):
        """ Return a list of an attribute for each scan's what group. """
        return np.array([self._hfile[s]['what'].attrs[attr]
                         for s in self._scans], dtype=dtype)

    # misc looping
    def moment_groups(self):
        """ Return a list of groups under scan0 where moments are stored. """
        return [k for k in self._hfile['/scan0'] if k.startswith('moment_')]

    def moment_names(self, scan0_groups):
        """ Return a list of moment names for a list of scan0 groups. """
        if hasattr(
            self._hfile['/scan0'][scan0_groups[0]].attrs['moment'], 'decode'):
            return [self._hfile['/scan0'][k].attrs['moment'].decode('utf-8') for
                    k in scan0_groups]
        else:
            return [self._hfile['/scan0'][k].attrs['moment'] for k in scan0_groups]

    def is_field_in_ray_header(self, field):
        """ True if field is present in ray_header, False otherwise. """
        return field in self._hfile[self._scans[0]]['ray_header'].dtype.names

    def ray_header(self, field, dtype):
        """ Return an array containing a ray_header field for each sweep. """
        data = np.empty((self.total_rays, ), dtype=dtype)
        for scan, start, end in zip(self._scans, self.start_ray, self.end_ray):
            data[start:end+1] = self._hfile[scan]['ray_header'][field]
        return data

    def moment_data(self, group, dtype):
        """ Read in moment data from all sweeps. """
        ngates = int(self._hfile['/scan0/how'].attrs['bin_count'])
        data = np.ma.zeros((self.total_rays, ngates), dtype=dtype)
        data[:] = np.ma.masked      # volume data initially all masked
        for scan, start, end in zip(self._scans, self.start_ray, self.end_ray):
            # read in sweep data if field exists in scan.
            if group in self._hfile[scan]:
                sweep_data = _get_gamic_sweep_data(self._hfile[scan][group])
                data[start:end+1, :sweep_data.shape[1]] = sweep_data[:]
        return data

    def sweep_expand(self, arr, dtype='float32'):
        """ Expand an sweep indexed array to be ray indexed """
        return np.repeat(arr, self.rays_per_sweep).astype(dtype)


def _get_gamic_sweep_data(group):
    """ Get GAMIC HDF5 sweep data from an HDF5 group. """
    dyn_range_min = group.attrs['dyn_range_min']
    dyn_range_max = group.attrs['dyn_range_max']
    raw_data = group[:]
    fmt = group.attrs['format']
    if hasattr(fmt, 'decode'):
        fmt = fmt.decode('UTF-8')
    if fmt == 'UV16':
        # unsigned 16-bit integer data, 0 indicates a masked value
        assert raw_data.dtype == np.uint16
        scale = (dyn_range_max - dyn_range_min) / 65535.
        offset = dyn_range_min
        sweep_data = np.ma.masked_array(
            raw_data * scale + offset, mask=(raw_data == 0), dtype='float32')
    elif fmt == 'UV8':
        # unsigned 8-bit integer data, 0 indicates a masked value
        assert raw_data.dtype == np.uint8
        scale = (dyn_range_max - dyn_range_min) / 255.
        offset = dyn_range_min
        sweep_data = np.ma.masked_array(
            raw_data * scale + offset, mask=(raw_data == 0), dtype='float32')
    elif fmt == 'F':
        assert raw_data.dtype.type == np.float32
        sweep_data = np.ma.masked_array(
            raw_data, mask=np.isnan(raw_data), dtype='float32')
    else:
        raise NotImplementedError('GAMIC data format: %s', fmt)
    return sweep_data



"""
Cython wrapper around the NASA TRMM RSL library.

"""

cimport pyart.io._rsl_h as _rsl_h


# the next line is required so that RSL_F_LIST and RSL_INVF_LIST can be
# properly wrapped as Cython does not export the typedef from _rsl_h
# define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION" [-W#warnings]
ctypedef unsigned short Range
import numpy as np
cimport numpy as np
from datetime import datetime, timedelta


_RSL_VERSION_STR = _rsl_h._RSL_VERSION_STR


cpdef copy_volume(_RslVolume volume):
    """
    copy_volume(volume)

    Return a copy of a _RslVolume object.

    Parameters
    ----------
    volume : _RslVolume
        _RslVolume object to create a copy of.

    Returns
    -------
    nvolume : _RslVolume
        Copy of volume.

    """
    volume_copy = _rsl_h.RSL_copy_volume(volume._Volume)
    rslvolume = _RslVolume()
    rslvolume.load(volume_copy)
    rslvolume._dealloc = 1
    return rslvolume


cpdef create_volume(
        np.ndarray[np.float32_t, ndim=2] arr,
        np.ndarray[np.int32_t, ndim=1] rays_per_sweep,
        int vol_num=1):
    """
    create_volume(arr, rays_per_sweep, vol_num=1)

    Create a _RslVolume object from a 2D float32 array.

    No headers parameters except nsweeps, nrays and nbins are not set in the
    resulting _RslVolume object.

    Parameters
    ----------
    arr : array, 2D, float32
        Two dimensional float32 array.
    rays_per_sweep: array, 1D, int32
        Array listing number of rays in each sweep.
    vol_num : int
        Volume number used to set f and invf in the header.  The default is
        for velocity fields.  Useful values are 0 for reflectivity and
        1 for velocity.

    Returns
    -------
    volumes : _RslVolume
        _RslVolume containing array data.

    """
    # these variables can be moved to the module level if used elsewhere
    cdef (float (*)(_rsl_h.Range)) * RSL_F_LIST = [
        _rsl_h.DZ_F, _rsl_h.VR_F, _rsl_h.SW_F, _rsl_h.CZ_F, _rsl_h.ZT_F,
        _rsl_h.DR_F, _rsl_h.LR_F, _rsl_h.ZD_F, _rsl_h.DM_F, _rsl_h.RH_F,
        _rsl_h.PH_F, _rsl_h.XZ_F, _rsl_h.CD_F, _rsl_h.MZ_F, _rsl_h.MD_F,
        _rsl_h.ZE_F, _rsl_h.VE_F, _rsl_h.KD_F, _rsl_h.TI_F, _rsl_h.DX_F,
        _rsl_h.CH_F, _rsl_h.AH_F, _rsl_h.CV_F, _rsl_h.AV_F, _rsl_h.SQ_F,
        _rsl_h.VS_F, _rsl_h.VL_F, _rsl_h.VG_F, _rsl_h.VT_F, _rsl_h.NP_F,
        _rsl_h.HC_F, _rsl_h.VC_F, _rsl_h.VR_F, _rsl_h.SW_F, _rsl_h.VR_F,
        _rsl_h.SW_F, _rsl_h.DZ_F, _rsl_h.CZ_F, _rsl_h.PH_F, _rsl_h.SD_F,
        _rsl_h.DZ_F, _rsl_h.DZ_F]

    cdef (_rsl_h.Range (*)(float)) * RSL_INVF_LIST = [
        _rsl_h.DZ_INVF, _rsl_h.VR_INVF, _rsl_h.SW_INVF, _rsl_h.CZ_INVF,
        _rsl_h.ZT_INVF, _rsl_h.DR_INVF, _rsl_h.LR_INVF, _rsl_h.ZD_INVF,
        _rsl_h.DM_INVF, _rsl_h.RH_INVF, _rsl_h.PH_INVF, _rsl_h.XZ_INVF,
        _rsl_h.CD_INVF, _rsl_h.MZ_INVF, _rsl_h.MD_INVF, _rsl_h.ZE_INVF,
        _rsl_h.VE_INVF, _rsl_h.KD_INVF, _rsl_h.TI_INVF, _rsl_h.DX_INVF,
        _rsl_h.CH_INVF, _rsl_h.AH_INVF, _rsl_h.CV_INVF, _rsl_h.AV_INVF,
        _rsl_h.SQ_INVF, _rsl_h.VS_INVF, _rsl_h.VL_INVF, _rsl_h.VG_INVF,
        _rsl_h.VT_INVF, _rsl_h.NP_INVF, _rsl_h.HC_INVF, _rsl_h.VC_INVF,
        _rsl_h.VR_INVF, _rsl_h.SW_INVF, _rsl_h.VR_INVF, _rsl_h.SW_INVF,
        _rsl_h.DZ_INVF, _rsl_h.CZ_INVF, _rsl_h.PH_INVF, _rsl_h.SD_INVF,
        _rsl_h.DZ_INVF, _rsl_h.DZ_INVF]

    cdef _rsl_h.Volume * volume
    cdef _rsl_h.Sweep * sweep
    cdef _rsl_h.Ray * ray
    cdef int ray_i

    nsweeps = rays_per_sweep.shape[0]
    nbins = arr.shape[1]

    volume = _rsl_h.RSL_new_volume(nsweeps)
    volume.h.nsweeps = nsweeps

    ray_index = 0
    for nsweep in range(nsweeps):
        nrays = rays_per_sweep[nsweep]
        sweep = _rsl_h.RSL_new_sweep(nrays)
        volume.sweep[nsweep] = sweep
        sweep.h.nrays = nrays
        for nray in range(nrays):
            ray = _rsl_h.RSL_new_ray(nbins)
            sweep.ray[nray] = ray
            ray.h.nbins = nbins
            ray.h.f = RSL_F_LIST[vol_num]
            ray.h.invf = RSL_INVF_LIST[vol_num]
            for nbin in range(nbins):
                ray.range[nbin] = ray.h.invf(arr[ray_index, nbin])
            ray_index += 1

    rslvolume = _RslVolume()
    rslvolume.load(volume)
    rslvolume._dealloc = 1
    return rslvolume


cpdef _label_volume(_RslVolume volume, radar):
    """
    _label_volume(volume, radar)

    Add labels for dealiasing to a _RslVolume object from a radar object.

    This function does not set all parameter in the _RslVolume suitable for
    writing out the volume, rather it set those parameters which must be set
    prior to using :py:func:`pyart.correct._fourdd_interface.fourdd_dealias`.

    Parameters
    ----------
    volume : _RslVolume
        Volume object to which parameters will be set as needed prior to
        dealiasing.  Object is manipulated in-place.
    radar : Radar
        Radar object from which parameters are taken.

    """

    cdef _rsl_h.Sweep * sweep
    cdef _rsl_h.Ray * ray
    cdef int ray_index

    vol = volume._Volume
    nsweeps = vol.h.nsweeps
    nbins = vol.sweep[0].ray[0].h.nbins

    gate_size = int(radar.range['meters_between_gates'])
    range_bin1 = int(radar.range['meters_to_center_of_first_gate'])
    if 'shape' in dir(radar.altitude['data']):
        if radar.altitude['data'].shape == ():
            alt = float(radar.altitude['data'])
        else:
            alt = radar.altitude['data'][0]
    else:
        alt = radar.altitude['data']

    nyq_vels = radar.instrument_parameters['nyquist_velocity']['data']
    azimuths = radar.azimuth['data']
    elevs = radar.elevation['data']

    # label the volume
    ray_index = 0
    vol.h.nsweeps = nsweeps
    for nsweep in range(nsweeps):
        sweep = vol.sweep[nsweep]
        nrays = sweep.h.nrays
        for nray in range(nrays):
            ray = sweep.ray[nray]
            ray.h.azimuth = azimuths[ray_index]
            ray.h.elev = elevs[ray_index]
            ray.h.nyq_vel = nyq_vels[ray_index]
            ray.h.range_bin1 = range_bin1
            ray.h.gate_size = gate_size
            ray.h.alt = alt
            ray_index += 1
    return


cdef class _RslRay:
    """
    A object for accessing RSL Ray data and header information

    This class should not be initalized from within Python.  _RslRay object are
    returned from the :py:func:`_RslSweep.get_ray` method.

    Attributes
    ----------
    month : int
        Date for this ray, month (1-12).
    day : int
        Date for this ray, day (1-31).
    year : int
        Date for this ray, year (eg. 1993).
    hour : int
        Time for this ray, hour (0-23).
    minute : int
        Time for this ray, minute (0-59).
    sec : float
        Time for this ray, second + fractor of second.
    unam_rng : float
        Unambiguous range in km.
    azimuth : float
        Mean azimuth angle in degrees for the ray, must be positive.
        0 for north, 90 for east, 270 for west.
    ray_num : int
        Ray number within a scan.
    elev : float
        Elevation angle in degrees.
    elev_num : int
        Elevation number within the volume scan.
    range_bin1 : int
        Range to first gate in meters.
    gate_size : int
        Gate size in meters.
    vel_res : float
        Doppler velocity resolution.
    sweep_rate : float
        Sweep rate, full sweeps / minute.
    prf : int
        Pulse repetition frequency in Hz.
    prf2 : int
        Second pulse repition frequenct for dual PRF data.
    azim_rate : float
        Sweep rate in degrees / second.
    fix_angle : float
        Elevation angle for the sweep in degrees.
    pitch : float
        Pitch angle.
    roll : float
        Roll angle.
    heading : float
        Heading.
    pitch_rate : float
        Pitch rate in angle / sec.
    roll_rate : float
        Roll rate in angle / sec.
    heading_rate : float
        Heading rate in angle / sec.
    lat : float
        Latitude in degrees.
    lon : float
        Longitude in degrees.
    alt : int
        Altitude in meters.
    rvs : float
        Radial velocity correction in meters / second.
    vel_east : float
        Platform velocity to the east in meters / second.  Negative values for
        velocity to the west.
    vel_north : float
        Platform velocity to the north in meters / second.  Negative values for
        velocity to the south.
    vel_up : float
        Platform velocity upward in meters / second.  Negative values for
        velocity downward.
    pulse_count : int
        Pulses used in a single dwell time.
    pulse_width : float
        Pulse width in microseconds.
    beam_width : float
        Beamwidth in degrees.
    frequency : float
        Carrier frequency in GHz.
    wavelength : float
        Wavelength in meters.
    nyq_vel : float
        Nyquist velocity in meters / second.
    nbins : int
        Number of array elements in ray data.

    """

    cdef _rsl_h.Ray * _Ray
    cdef int _dealloc

    def __dealloc__(self):
        if self._dealloc == 1:
            _rsl_h.RSL_free_ray(self._Ray)

    cdef load(self, _rsl_h.Ray * Ray):
        """ Load the _RslRay object, must be called after creation. """
        if Ray is NULL:
            raise ValueError('cannot load _RslRay with NULL')
        self._Ray = Ray
        self._dealloc = 0

    def get_datetime(self):
        """
        get_datetime()

        Return a datetime describing the date and time of the ray.
        """
        s = self
        full_seconds, fractional_seconds = divmod(s.sec, 1)
        microseconds = int(fractional_seconds * 1e6)
        # Some UF writers incorrectly specify midnight as 24:00:00 rather
        # than 00:00:00.  Handle this case explicitly
        if s.hour == 24:
            s.hour = 23
            s.minute = 0
            full_seconds = 0
            dt = datetime(s.year, s.month, s.day, s.hour, s.minute,
                          full_seconds, microseconds)
            return dt + timedelta(seconds=1)
        return datetime(s.year, s.month, s.day, s.hour, s.minute,
                        int(full_seconds), microseconds)

    def get_data(self):
        """
        get_data()

        Return the one-dimensional data contained in the ray.
        """
        cdef _rsl_h.Range raw
        cdef np.ndarray[np.float32_t, ndim = 1] data

        shape = (self._Ray.h.nbins)
        data = np.zeros(shape, dtype='float32') + 1.31072000e+05
        for nbin in range(self._Ray.h.nbins):
            raw = self._Ray.range[nbin]
            data[nbin] = self._Ray.h.f(raw)
        return data

    # header properties mapped to class attributes.
    property month:
        def __get__(self):
            return self._Ray.h.month

        def __set__(self, int month):
            self._Ray.h.month = month

    property day:
        def __get__(self):
            return self._Ray.h.day

        def __set__(self, int day):
            self._Ray.h.day = day

    property year:
        def __get__(self):
            return self._Ray.h.year

        def __set__(self, int year):
            self._Ray.h.year = year

    property hour:
        def __get__(self):
            return self._Ray.h.hour

        def __set__(self, int hour):
            self._Ray.h.hour = hour

    property minute:
        def __get__(self):
            return self._Ray.h.minute

        def __set__(self, int minute):
            self._Ray.h.minute = minute

    property sec:
        def __get__(self):
            return self._Ray.h.sec

        def __set__(self, float sec):
            self._Ray.h.sec = sec

    property unam_rng:
        def __get__(self):
            return self._Ray.h.unam_rng

        def __set__(self, float unam_rng):
            self._Ray.h.unam_rng = unam_rng

    property azimuth:
        def __get__(self):
            return self._Ray.h.azimuth

        def __set__(self, float azimuth):
            self._Ray.h.azimuth = azimuth

    property ray_num:
        def __get__(self):
            return self._Ray.h.ray_num

        def __set__(self, int ray_num):
            self._Ray.h.ray_num = ray_num

    property elev:
        def __get__(self):
            return self._Ray.h.elev

        def __set__(self, float elev):
            self._Ray.h.elev = elev

    property elev_num:
        def __get__(self):
            return self._Ray.h.elev_num

        def __set__(self, int elev_num):
            self._Ray.h.elev_num = elev_num

    property range_bin1:
        def __get__(self):
            return self._Ray.h.range_bin1

        def __set__(self, int range_bin1):
            self._Ray.h.range_bin1 = range_bin1

    property gate_size:
        def __get__(self):
            return self._Ray.h.gate_size

        def __set__(self, int gate_size):
            self._Ray.h.gate_size = gate_size

    property vel_res:
        def __get__(self):
            return self._Ray.h.vel_res

        def __set__(self, float vel_res):
            self._Ray.h.vel_res = vel_res

    property sweep_rate:
        def __get__(self):
            return self._Ray.h.sweep_rate

        def __set__(self, float sweep_rate):
            self._Ray.h.sweep_rate = sweep_rate

    property prf:
        def __get__(self):
            return self._Ray.h.prf

        def __set__(self, int prf):
            self._Ray.h.prf = prf

    property prf2:
        def __get__(self):
            return self._Ray.h.prf2

        def __set__(self, int prf2):
            self._Ray.h.prf2 = prf2

    property azim_rate:
        def __get__(self):
            return self._Ray.h.azim_rate

        def __set__(self, float azim_rate):
            self._Ray.h.azim_rate = azim_rate

    property fix_angle:
        def __get__(self):
            return self._Ray.h.fix_angle

        def __set__(self, float fix_angle):
            self._Ray.h.fix_angle = fix_angle

    property pitch:
        def __get__(self):
            return self._Ray.h.pitch

        def __set__(self, float pitch):
            self._Ray.h.pitch = pitch

    property roll:
        def __get__(self):
            return self._Ray.h.roll

        def __set__(self, float roll):
            self._Ray.h.roll = roll

    property heading:
        def __get__(self):
            return self._Ray.h.heading

        def __set__(self, float heading):
            self._Ray.h.heading = heading

    property pitch_rate:
        def __get__(self):
            return self._Ray.h.pitch_rate

        def __set__(self, float pitch_rate):
            self._Ray.h.pitch_rate = pitch_rate

    property roll_rate:
        def __get__(self):
            return self._Ray.h.roll_rate

        def __set__(self, float roll_rate):
            self._Ray.h.roll_rate = roll_rate

    property heading_rate:
        def __get__(self):
            return self._Ray.h.heading_rate

        def __set__(self, float heading_rate):
            self._Ray.h.heading_rate = heading_rate

    property lat:
        def __get__(self):
            return self._Ray.h.lat

        def __set__(self, float lat):
            self._Ray.h.lat = lat

    property lon:
        def __get__(self):
            return self._Ray.h.lon

        def __set__(self, float lon):
            self._Ray.h.lon = lon

    property alt:
        def __get__(self):
            return self._Ray.h.alt

        def __set__(self, int alt):
            self._Ray.h.alt = alt

    property rvc:
        def __get__(self):
            return self._Ray.h.rvc

        def __set__(self, float rvc):
            self._Ray.h.rvc = rvc

    property vel_east:
        def __get__(self):
            return self._Ray.h.vel_east

        def __set__(self, float vel_east):
            self._Ray.h.vel_east = vel_east

    property vel_north:
        def __get__(self):
            return self._Ray.h.vel_north

        def __set__(self, float vel_north):
            self._Ray.h.vel_north = vel_north

    property vel_up:
        def __get__(self):
            return self._Ray.h.vel_up

        def __set__(self, float vel_up):
            self._Ray.h.vel_up = vel_up

    property pulse_count:
        def __get__(self):
            return self._Ray.h.pulse_count

        def __set__(self, int pulse_count):
            self._Ray.h.pulse_count = pulse_count

    property pulse_width:
        def __get__(self):
            return self._Ray.h.pulse_width

        def __set__(self, float pulse_width):
            self._Ray.h.pulse_width = pulse_width

    property beam_width:
        def __get__(self):
            return self._Ray.h.beam_width

        def __set__(self, float beam_width):
            self._Ray.h.beam_width = beam_width

    property frequency:
        def __get__(self):
            return self._Ray.h.frequency

        def __set__(self, float frequency):
            self._Ray.h.frequency = frequency

    property wavelength:
        def __get__(self):
            return self._Ray.h.wavelength

        def __set__(self, float wavelength):
            self._Ray.h.wavelength = wavelength

    property nyq_vel:
        def __get__(self):
            return self._Ray.h.nyq_vel

        def __set__(self, float nyq_vel):
            self._Ray.h.nyq_vel = nyq_vel

    property nbins:
        def __get__(self):
            return self._Ray.h.nbins

        def __set__(self, int nbins):
            self._Ray.h.nbins = nbins


cdef class _RslSweep:
    """
    A object for accessing RSL Sweep data and header information.

    This class should not be initalized from within Python.  _RslSweep objects
    are returned from the :py:func:`_RslVolume.get_sweep` method.

    Attributes
    ----------
    sweep_num : int
        Interger sweep number.
    elev : float
        Mean elevation angle for thr sweep.  -999.0 for RHI sweeps.
    azimuth : float
        Azumuth for the sweep.  -999.0 for PPI scans.
    beam_width : float
        Beam width in degrees.  Can also be found in _RslRay objects.
    vert_half_bw : float
        Vertical beam width divided by 2.
    horz_half_bw : float
        Horizontal beam width divided by 2.
    nrays : int
        Number of rays in the sweep.

    """

    cdef _rsl_h.Sweep * _Sweep
    cdef int _dealloc

    def __dealloc__(self):
        if self._dealloc == 1:
            _rsl_h.RSL_free_sweep(self._Sweep)

    cdef load(self, _rsl_h.Sweep * Sweep):
        """ Load the _RslSweep object, must be called after creation. """
        if Sweep is NULL:
            raise ValueError("cannot load _RslSweep with NULL")
        self._Sweep = Sweep
        self._dealloc = 0

    def get_ray(self, int ray_number):
        """
        get_ray(ray_number)

        Return a _RslRay for a given ray.

        Parameters
        ----------
        ray_number : int
            Ray number to retrieve

        Returns
        -------
        ray : _RslRay
            _RslRay object containing the requested ray.

        """
        if ray_number < 0 or ray_number >= self._Sweep.h.nrays:
            raise ValueError('invalid ray_number')
        rslray = _RslRay()
        rslray.load(self._Sweep.ray[ray_number])
        return rslray

    def get_data(self):
        """
        get_data()

        Return the two-dimensional data contained in the sweep.

        If a given ray has few bins than the first ray, the missing bins
        will be filled with 131072.0
        """
        cdef _rsl_h.Range raw
        cdef _rsl_h.Ray * ray
        cdef np.ndarray[np.float32_t, ndim = 2] data

        sweep = self._Sweep
        nrays = sweep.h.nrays
        nbins = sweep.ray[0].h.nbins

        shape = (nrays, nbins)
        data = np.zeros(shape, dtype='float32') + 1.31072000e+05
        for nray in range(nrays):
            ray = sweep.ray[nray]
            assert ray is not NULL
            nbins = ray.h.nbins
            for nbin in range(nbins):
                raw = ray.range[nbin]
                data[nray, nbin] = ray.h.f(raw)
        return data

    # header properties mapped to class attributes.
    property sweep_num:
        def __get__(self):
            return self._Sweep.h.sweep_num

        def __set__(self, int sweep_num):
            self._Sweep.h.sweep_num = sweep_num

    property elev:
        def __get__(self):
            return self._Sweep.h.elev

        def __set__(self, float elev):
            self._Sweep.h.elev = elev

    property azimuth:
        def __get__(self):
            return self._Sweep.h.azimuth

        def __set__(self, float azimuth):
            self._Sweep.h.azimuth = azimuth

    property beam_width:
        def __get__(self):
            return self._Sweep.h.beam_width

        def __set__(self, float beam_width):
            self._Sweep.h.beam_width = beam_width

    property vert_half_bw:
        def __get__(self):
            return self._Sweep.h.vert_half_bw

        def __set__(self, float vert_half_bw):
            self._Sweep.h.vert_half_bw = vert_half_bw

    property horz_half_bw:
        def __get__(self):
            return self._Sweep.h.horz_half_bw

        def __set__(self, float horz_half_bw):
            self._Sweep.h.horz_half_bw = horz_half_bw

    property nrays:
        def __get__(self):
            return self._Sweep.h.nrays

        def __set__(self, int nrays):
            self._Sweep.h.nrays = nrays


cdef class _RslVolume:
    """
    A object for accessing RSL Volume data and header information.

    This class should not be initalized from within Python.  _RslVolume
    objects are returned from the :py:func:`RslFile.get_volume` and other
    functions/methods.

    Attributes
    ----------
    nsweeps : int
        Sweep number.
    calibr_const : float
        Calibration constant.

    """

    def __dealloc__(self):
        if self._dealloc == 1:
            _rsl_h.RSL_free_volume(self._Volume)

    cdef load(self, _rsl_h.Volume * Volume):
        """ Load the _RslVolume object, must be called after creation. """
        if Volume is NULL:
            raise ValueError('cannot load _RslVolume with NULL')
        self._Volume = Volume
        self._dealloc = 0

    def total_rays(self):
        """
        total_rays()

        Return the total number of rays present in all sweeps of the volume.
        """
        return np.sum(self.get_nray_array())

    def get_nray_array(self):
        """
        get_nray_array()

        Return an array of the number of rays for each sweep.
        """
        cdef _rsl_h.Sweep * sweep
        nrays = np.empty((self.nsweeps), dtype='int32')
        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            nrays[i] = sweep.h.nrays
        return nrays

    def get_sweep(self, int sweep_number):
        """
        get_sweep(sweep_numer)

        Return a _RslSweep for a given sweep number.

        Parameters
        ----------
        sweep_number : int
            Sweep number to retrieve

        Returns
        -------
        sweep : _RslSweep
            _RslSweep object containing the requested sweep.

        """
        if sweep_number < 0 or sweep_number >= self._Volume.h.nsweeps:
            raise ValueError('invalid sweep_number')

        rslsweep = _RslSweep()
        rslsweep.load(self._Volume.sweep[sweep_number])
        return rslsweep

    def get_azimuth_and_elev_array(self):
        """
        get_azimuth_and_elev_array()

        Return azimuth and elevation array for each sweep and ray.
        """
        cdef int nrays = self._Volume.sweep[0].h.nrays
        cdef int ray_count
        cdef _rsl_h.Sweep * sweep
        cdef _rsl_h.Ray * ray

        # create empty azimuth and elev output arrays
        total_rays = self.total_rays()
        azimuth = np.empty([total_rays], dtype='float32')
        elev = np.empty([total_rays], dtype='float32')

        # loop over the sweeps and rays storing azimuth and elev
        ray_count = 0
        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            nrays = sweep.h.nrays
            for j in range(nrays):
                ray = sweep.ray[j]
                assert ray is not NULL
                azimuth[ray_count + j] = ray.h.azimuth
                elev[ray_count + j] = ray.h.elev
            ray_count += nrays
        return azimuth, elev

    def get_sweep_fix_angles(self):
        """
        get_sweep_fix_angles()

        Return array of fix angle for each sweep.

        Angles determined from the first ray in each sweep.
        """
        cdef _rsl_h.Sweep * sweep
        cdef _rsl_h.Ray * ray
        fix_angles = np.empty((self.nsweeps), dtype='float32')
        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            ray = sweep.ray[0]
            assert ray is not NULL
            fix_angles[i] = ray.h.fix_angle
        return fix_angles

    def get_sweep_azimuths(self):
        """
        get_sweep_azimuths()

        Return azimuth array for each sweep.
        """
        cdef _rsl_h.Sweep * sweep
        azimuth = np.empty((self.nsweeps), dtype='float32')
        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            azimuth[i] = sweep.h.azimuth
        return azimuth

    def get_sweep_elevs(self):
        """
        get_sweep_elevs()

        Return elevation array for each sweep.
        """
        cdef _rsl_h.Sweep * sweep
        elev = np.empty((self.nsweeps), dtype='float32')
        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            elev[i] = sweep.h.elev
        return elev

    def get_instr_params(self):
        """
        get_instr_params()

        Return instrumental parameter for the volume.

        Returns
        -------
        pm_data : array, (nsweeps)
            Array of prt modes.
        nv_data : array, (total_rays)
            Array of nyquist velocities.
        pr_data : array, (total_rays)
            Array of pulse repetition frequency in Hz.
        ur_data : array, (total_rays)
            Array of unambiguous ranges, in km.

        """
        cdef int nrays = self._Volume.sweep[0].h.nrays
        cdef int ray_count
        cdef _rsl_h.Sweep * sweep
        cdef _rsl_h.Ray * ray

        # calculate the total number of rays in the volume

        # initalize empty instrument parameter arrays
        total_rays = self.total_rays()
        nyq_vel = self._Volume.sweep[0].ray[0].h.nyq_vel
        valid_nyq_vel = abs(nyq_vel) > 0.1
        pm_data = np.empty(self.nsweeps, dtype='|S24')
        nv_data = np.empty((total_rays), dtype='float32')
        pr_data = np.empty((total_rays), dtype='float32')
        ur_data = np.empty((total_rays), dtype='float32')

        # loop over sweeps and rays storing instrument parameters
        ray_count = 0
        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            nrays = sweep.h.nrays
            for j in range(nrays):
                ray = sweep.ray[j]
                assert ray is not NULL
                if j == 0:
                    pm_data[i] = self._prtmode(ray.h)

                if valid_nyq_vel:
                    nv_data[ray_count + j] = ray.h.nyq_vel
                else:
                    nv_data[ray_count + j] = (ray.h.wavelength *
                                              ray.h.prf / 4.0)

                if ray.h.prf == 0:
                    pr_data[ray_count + j] = -999.
                else:
                    pr_data[ray_count + j] = 1. / ray.h.prf
                ur_data[ray_count + j] = ray.h.unam_rng * 1000.0
            ray_count += nrays
        return pm_data, nv_data, pr_data, ur_data

    def get_data(self):
        """
        get_data()

        Return the two-dimensional data contained in the volume.

        If a given ray has few bins than the first ray, the missing bins
        will be filled with 131072.0
        """
        cdef _rsl_h.Range raw
        cdef _rsl_h.Sweep * sweep
        cdef _rsl_h.Ray * ray
        cdef int ray_count, nsweeps, nrays, nbins, nray, nsweep, nbin
        cdef np.ndarray[np.float32_t, ndim = 2] data

        vol = self._Volume
        nbins = vol.sweep[0].ray[0].h.nbins
        total_rays = self.total_rays()
        shape = (total_rays, nbins)
        data = np.zeros(shape, dtype='float32') + 1.31072000e+05

        ray_count = 0
        nsweeps = vol.h.nsweeps
        for nsweep in range(nsweeps):
            sweep = vol.sweep[nsweep]
            assert sweep is not NULL
            nrays = sweep.h.nrays
            for nray in range(nrays):
                ray = sweep.ray[nray]
                assert ray is not NULL
                nbins = ray.h.nbins
                for nbin in range(nbins):
                    raw = ray.range[nbin]
                    data[ray_count + nray, nbin] = ray.h.f(raw)
            ray_count += nrays
        return data

    def is_range_bins_uniform(self):
        """
        is_range_bins_uniform()

        Return True is the locations of the range bin are identical for all
        rays, False if locations change in one or more rays.
        """
        cdef int nrays = self._Volume.sweep[0].h.nrays
        cdef _rsl_h.Sweep * sweep
        cdef _rsl_h.Ray * ray

        # loop over the sweeps and rays checking that the gate_size and
        # range_bin1 are the same as the that in the first ray
        sweep = self._Volume.sweep[0]
        assert sweep is not NULL
        ray = sweep.ray[0]
        assert ray is not NULL
        ref_gate_size = ray.h.gate_size
        ref_range_bin1 = ray.h.range_bin1

        for i in range(self.nsweeps):
            sweep = self._Volume.sweep[i]
            assert sweep is not NULL
            nrays = sweep.h.nrays
            for j in range(nrays):
                ray = sweep.ray[j]
                assert ray is not NULL
                if ray.h.gate_size != ref_gate_size:
                    return False
                if ray.h.range_bin1 != ref_range_bin1:
                    return False
        return True

    cdef _prtmode(self, _rsl_h.Ray_header h):
        """ Return the prt mode of a given Ray header. """
        # TODO need to add additional logic here
        if h.prf2 != h.prf:
            mode = 'dual                    '
        else:
            mode = 'fixed                   '
        return mode

    # header properties mapped to class attributes.
    property nsweeps:
        def __get__(self):
            return self._Volume.h.nsweeps

        def __set__(self, int nsweeps):
            self._Volume.h.nsweeps = nsweeps

    property calibr_const:
        def __get__(self):
            return self._Volume.h.calibr_const

        def __set__(self, float calibr_const):
            self._Volume.h.calibr_const = calibr_const


cdef class RslFile:
    """
    RslFile(filename)

    A object for accessing Radar data and parameter using the RSL library.

    Parameters
    ----------
    filename : str
        Radar file to read.

    Attributes
    ----------
    month : int
        Date, month (1-12).
    day : int
        Date, day (1-31).
    year : int
        Date, year (eg. 1993).
    hour : int
        Time, hour (0-23).
    minute : int
        Time, minute (0-59).
    sec : float
        Time, second + fractions of second.
    nvolumes : int
        Number of volume slots in the file.
    number : int
        Arbitrary number for this radar site.
    latd, latm, lats : int
        Latitude degrees, minutes and seconds for the site.
    lond, lonm, lons : int
        Longitude degrees, minutes and seconds for the site.
    height : int
        Height of site in meters above sea level.
    spulse : int
        Length of short pulse in ns.
    lpulse : int
        Length of long pulse in ns.
    scan_mode : int
        Scan mode, 0 for PPI, 1 for RHI.
    vcp : int
        Volume coverage pattern, WSR-88D only.

    """
    cdef _rsl_h.Radar * _Radar
    cdef _rsl_h.Volume * _Volume
    cdef _rsl_h.Sweep * _Sweep
    cdef _rsl_h.Ray * _Ray

    def __cinit__(self, filename, radar_format=None, callid=None):
        """ Initalize the _RslFile object. """
        if radar_format == 'wsr88d':
            if callid is None:
                raise ValueError('callid must be provided.')
            self._Radar = _rsl_h.RSL_wsr88d_to_radar(filename, callid)
        elif radar_format is None:
            self._Radar = _rsl_h.RSL_anyformat_to_radar(filename)
        else:
            raise ValueError('invalid radar_format:', radar_format)
        if self._Radar is NULL:
            raise IOError('file cannot be read.')

    def __dealloc__(self):
        """ Free memory used by object. """
        _rsl_h.RSL_free_radar(self._Radar)

    def get_volume(self, int volume_number):
        """
        get_volume(volume_number)

        Return a _RslVolume for a given volume number.

        Parameters
        ----------
        volume_number : int
            Volume number to retrieve

        Returns
        -------
        volume : _RslVolume
            _RslVolume object containing requested volume.

        """
        if volume_number < 0 or volume_number >= self._Radar.h.nvolumes:
            raise ValueError('invalid volume_number')

        rslvolume = _RslVolume()
        rslvolume.load(self._Radar.v[volume_number])
        return rslvolume

    def available_moments(self):
        """
        available_moments()

        Return a list of available volume moments.
        """
        av = []
        for i in range(self._Radar.h.nvolumes):
            if self._Radar.v[i] is not NULL:
                av.append(i)
        return av

    def get_radar_header(self):
        """
        get_radar_headers()

        Return a dictionary of radar header parameters.
        """
        return self._Radar.h

    def get_volume_array(self, int volume_num):
        """
        get_volume_array(volume_number)

        Return the three-dimensional data contained in a given volume.

        Parameters
        ----------
        volume_number : int

        Returns
        -------
        volume : array (nsweep, nrays, nbins), float32
            Array containing  data for the given volume.

        """
        return self.get_volume(volume_num).get_data()

    # header properties mapped to class attributes.
    property month:
        def __get__(self):
            return self._Radar.h.month

        def __set__(self, int month):
            self._Radar.h.month = month

    property day:
        def __get__(self):
            return self._Radar.h.day

        def __set__(self, int day):
            self._Radar.h.day = day

    property year:
        def __get__(self):
            return self._Radar.h.year

        def __set__(self, int year):
            self._Radar.h.year = year

    property hour:
        def __get__(self):
            return self._Radar.h.hour

        def __set__(self, int hour):
            self._Radar.h.hour = hour

    property minute:
        def __get__(self):
            return self._Radar.h.minute

        def __set__(self, int minute):
            self._Radar.h.minute = minute

    property sec:
        def __get__(self):
            return self._Radar.h.sec

        def __set__(self, float sec):
            self._Radar.h.sec = sec

    property nvolumes:
        def __get__(self):
            return self._Radar.h.nvolumes

        def __set__(self, int nvolumes):
            self._Radar.h.nvolumes = nvolumes

    property number:
        def __get__(self):
            return self._Radar.h.number

        def __set__(self, int number):
            self._Radar.h.number = number

    property latd:
        def __get__(self):
            return self._Radar.h.latd

        def __set__(self, int latd):
            self._Radar.h.latd = latd

    property latm:
        def __get__(self):
            return self._Radar.h.latm

        def __set__(self, int latm):
            self._Radar.h.latm = latm

    property lats:
        def __get__(self):
            return self._Radar.h.lats

        def __set__(self, int lats):
            self._Radar.h.lats = lats

    property lond:
        def __get__(self):
            return self._Radar.h.lond

        def __set__(self, int lond):
            self._Radar.h.lond = lond

    property lonm:
        def __get__(self):
            return self._Radar.h.lonm

        def __set__(self, int lonm):
            self._Radar.h.lonm = lonm

    property lons:
        def __get__(self):
            return self._Radar.h.lons

        def __set__(self, int lons):
            self._Radar.h.lons = lons

    property height:
        def __get__(self):
            return self._Radar.h.height

        def __set__(self, int height):
            self._Radar.h.height = height

    property spulse:
        def __get__(self):
            return self._Radar.h.spulse

        def __set__(self, int spulse):
            self._Radar.h.spulse = spulse

    property lpulse:
        def __get__(self):
            return self._Radar.h.lpulse

        def __set__(self, int lpulse):
            self._Radar.h.lpulse = lpulse

    property scan_mode:
        def __get__(self):
            return self._Radar.h.scan_mode

        def __set__(self, int scan_mode):
            self._Radar.h.scan_mode = scan_mode

    property vcp:
        def __get__(self):
            return self._Radar.h.vcp

        def __set__(self, int vcp):
            self._Radar.h.vcp = vcp



"""
A dictionary-like class supporting lazy loading of specified keys.

"""

from collections.abc import MutableMapping
import itertools


class LazyLoadDict(MutableMapping):
    """
    A dictionary-like class supporting lazy loading of specified keys.

    Keys which are lazy loaded are specified using the set_lazy method.
    The callable object which produces the specified key is provided as the
    second argument to this method. This object gets called when the value
    of the key is loaded. After this initial call the results is cached
    in the traditional dictionary which is used for supplemental access to
    this key.

    Testing for keys in this dictionary using the "key in d" syntax will
    result in the loading of a lazy key, use "key in d.keys()" to prevent
    this evaluation.

    The comparison methods, __cmp__, __ge__, __gt__, __le__, __lt__, __ne__,
    nor the view methods, viewitems, viewkeys, viewvalues, are implemented.
    Neither is the the fromkeys method.

    Parameters
    ----------
    dic : dict
        Dictionary containing key, value pairs which will be stored and
        evaluated traditionally. This dictionary referenced not copied into
        the LazyLoadDictionary and hence changed to this dictionary may change
        the original. If this behavior is not desired copy dic in the
        initalization.

    Examples
    --------
    >>> d = LazyLoadDict({'key1': 'value1', 'key2': 'value2'})
    >>> d.keys()
    ['key2', 'key1']
    >>> lazy_func = lambda : 999
    >>> d.set_lazy('lazykey1', lazy_func)
    >>> d.keys()
    ['key2', 'key1', 'lazykey1']
    >>> d['lazykey1']
    999

    """
    def __init__(self, dic):
        """ initalize. """
        self._dic = dic
        self._lazyload = {}

    # abstract methods
    def __setitem__(self, key, value):
        """ Set a key which will not be stored and evaluated traditionally. """
        self._dic[key] = value
        if key in self._lazyload:
            del self._lazyload[key]

    def __getitem__(self, key):
        """ Get the value of a key, evaluating a lazy key if needed. """
        if key in self._lazyload:
            value = self._lazyload[key]()
            self._dic[key] = value
            del self._lazyload[key]
        return self._dic[key]

    def __delitem__(self, key):
        """ Remove a lazy or traditional key from the dictionary. """
        if key in self._lazyload:
            del self._lazyload[key]
        else:
            del self._dic[key]

    def __iter__(self):
        """ Iterate over all lazy and traditional keys. """
        return itertools.chain(self._dic.copy(), self._lazyload.copy())

    def __len__(self):
        """ Return the number of traditional and lazy keys. """
        return len(self._dic) + len(self._lazyload)

    # additional class to mimic dict behavior
    def __str__(self):
        """ Return a string representation of the object. """
        if len(self._dic) == 0 or len(self._lazyload) == 0:
            seperator = ''
        else:
            seperator = ', '
        lazy_reprs = [(repr(k), repr(v)) for k, v in self._lazyload.items()]
        lazy_strs = ['%s: LazyLoad(%s)' % r for r in lazy_reprs]
        lazy_str = ", ".join(lazy_strs) + '}'
        return str(self._dic)[:-1] + seperator + lazy_str

    def has_key(self, key):
        """ True if dictionary has key, else False. """
        return key in self

    def copy(self):
        """
        Return a copy of the dictionary.

        Lazy keys are not evaluated in the original or copied dictionary.
        """
        dic = self.__class__(self._dic.copy())
        # load all lazy keys into the copy
        for key, value_callable in self._lazyload.items():
            dic.set_lazy(key, value_callable)
        return dic

    # lazy dictionary specific methods
    def set_lazy(self, key, value_callable):
        """ Set a lazy key to load from a callable object. """
        if key in self._dic:
            del self._dic[key]
        self._lazyload[key] = value_callable



"""
Routines for Ridgeline Instruments RXM-25 formatted NetCDF files.

"""

import datetime

try:
    import pytz
    _PYTZ_AVAILABLE = True
except ImportError:
    _PYTZ_AVAILABLE = False

import netCDF4
import numpy as np

import pyart
from ..config import get_metadata
from ..core.radar import Radar
from ..exceptions import MissingOptionalDependency
from ..testing import make_empty_ppi_radar

def read_rxm25(filename, cfradial_outfile=None, heading=None):
    """
    Read in Ridgeline Instruments RXM-25 formatted NetCDF data.

    Parameters
    ----------
    filename : str
        Name of Ridgeline Instruments (RLI) RXM-25 formatted NetCDF file 
        from which to read data.
    cfradial_outfile : str, optional
        If file is to be converted to CF-Radial format, specify the output 
        filename here.
    heading : float, optional
        If a heading offset exists, enter it here (in degrees).

    Returns
    -------
    radar : Radar
        Radar object.

    """
    if not _PYTZ_AVAILABLE:
        raise MissingOptionalDependency(
            "Pytz is required to use read_rxm25 but is " +
            "not installed")

    data  = netCDF4.Dataset(filename, 'r')

    ngates = data.dimensions['Gate'].size
    rays_per_sweep = data.dimensions['Radial'].size
    radar = make_empty_ppi_radar(ngates, rays_per_sweep, 1)

    # Time needs to be converted from nss1970 to nss1989 and added to
    # Radar object.
    nineteen89 = datetime.datetime(1989, 1, 1, 0, 0, 1, tzinfo = pytz.utc)
    baseTime = np.array(
        [datetime.datetime.fromtimestamp(
            t, tz=pytz.UTC) for t in data.variables['Time'][:]])
    radar.time['data'] = np.array(
        [t.total_seconds() for t in baseTime - nineteen89])

    if heading is not None:
        radar.heading = heading
        radar.azimuth['data'] = np.mod(
            data['Azimuth'][:] - radar.heading, 360.)
    else:
        radar.azimuth['data'] = data['Azimuth'][:]

    radar.longitude['data'] = np.array([data.Longitude], dtype='float64')
    radar.latitude['data'] = np.array([data.Latitude], dtype='float64')
    radar.elevation['data'] = data['Elevation'][:]
    radar.altitude['data'] = np.array([data.Height], dtype='float64')

    fixed_agl_data = np.empty((1, ), dtype='float32')
    fixed_agl_data[:] = np.mean(
        radar.elevation['data'][:rays_per_sweep])

    radar.fixed_angle['data'] = fixed_agl_data

    radar.range['data'] = np.linspace(
        data['StartRange'][0]/1000,
        (ngates - 1)*data['GateWidth'][0]/1000 + data['StartRange'][0]/1000,
        ngates)

    ref = data['Reflectivity'][:]
    norm_pow = data['NormalizedCoherentPower'][:]
    spec_w = data['SpectralWidth'][:]
    vel = data['Velocity'][:]
    corr_ref = data['CorrectedReflectivity'][:]
    diff_ref = data['DifferentialReflectivity'][:]
    diff_phase = data['DifferentialPhase'][:]
    spec_phase = data['SpecificPhase'][:]
    corr_diff_ref = data['CorrectedDifferentialReflectivity'][:]
    sig_noise = data['SignalToNoiseRatio'][:]
    rain_rate = data['RainfallRate'][:]
    cross_ra = data['CrossPolCorrelation'][:]

    fields = {
        'reflectivity': get_metadata('reflectivity'),
        'normalized_coherent_power': get_metadata('normalized_coherent_power'),
        'spectral_width': get_metadata('spectral_width'),
        'velocity': get_metadata('velocity'),
        'corrected_reflectivity': get_metadata('correct_reflectivity'),
        'differential_reflectivity': get_metadata('differential_reflectivity'),
        'differential_phase': get_metadata('differential_phase'),
        'specific_differential_phase': get_metadata('specific_differential_phase'),
        'corrected_differential_reflectivity': get_metadata('corrected_differential_reflectivity'),
        'signal_to_noise_ratio': get_metadata('signal_to_noise_ratio'),
        'rain_rate': get_metadata('rain_rate'),
        'cross_correlation_ratio': get_metadata('cross_correlation_ratio')}
    
    radar.fields = fields
    radar.fields['reflectivity']['data'] = ref
    radar.fields['normalized_coherent_power']['data'] = norm_pow
    radar.fields['spectral_width']['data'] = spec_w
    radar.fields['velocity']['data'] = vel
    radar.fields['corrected_reflectivity']['data'] = corr_ref
    radar.fields['differential_reflectivity']['data'] = diff_ref
    radar.fields['differential_phase']['data'] = diff_phase
    radar.fields['specific_differential_phase']['data'] = spec_phase
    radar.fields['corrected_differential_reflectivity']['data'] = corr_diff_ref
    radar.fields['signal_to_noise_ratio']['data'] = sig_noise
    radar.fields['rain_rate']['data'] = rain_rate
    radar.fields['cross_correlation_ratio']['data'] = cross_ra

    radar.metadata['instrument_name'] = 'RXM-25'

    if cfradial_outfile is not None:
        pyart.io.write_cfradial(
            cfradial_outfile, radar, arm_time_variables=True)

    return radar



"""
Reading files using Radx to first convert the file to Cf.Radial format

"""

import os
import tempfile
import subprocess

from ..io.cfradial import read_cfradial
from ..io.common import _test_arguments


def read_radx(filename, radx_dir=None, **kwargs):
    """
    Read a file by first converting it to Cf/Radial using RadxConvert.

    Parameters
    ----------
    filename : str
        Name of file to read using RadxConvert.

    radx_dir : str, optional
        path to the radx install

    Returns
    -------
    radar : Radar
        Radar object.

    """
    # test for non empty kwargs
    _test_arguments(kwargs)
    if radx_dir is not None:
        executable = os.path.join(radx_dir, 'RadxConvert')
    else:
        executable = 'RadxConvert'

    tmpfile = tempfile.mkstemp(suffix='.nc', dir='.')[1]
    head, tail = os.path.split(tmpfile)
    try:
        subprocess.check_call(
            [executable, '-const_ngates',
             '-outdir', head, '-outname', tail, '-f', filename])
        if not os.path.isfile(tmpfile):
            raise IOError(
                'RadxConvert failed to create a file, upgrading to the '
                ' latest version of Radx may be necessary.')
        radar = read_cfradial(tmpfile)
    finally:
        os.remove(tmpfile)
    return radar



""" Unit Tests for Py-ART's util/columnsect.py module. """

import numpy as np
import pyart

# read in example file
radar = pyart.io.read_nexrad_archive(pyart.testing.NEXRAD_ARCHIVE_MSG31_FILE)


def test_get_azimuth():
    # test to make sure azimuth is correct to Everett, WA
    azimuth = pyart.util.columnsect.for_azimuth(radar.latitude['data'][0],
                                                47.97,
                                                radar.longitude['data'][0],
                                                -122.20)
    test_azi = abs(np.around(azimuth, 2) - 138.57)
    assert (test_azi < 0.001)


def test_sphere_distance():
    # test to make sure sphere distance is correct to Everett, WA
    distance = pyart.util.columnsect.sphere_distance(radar.latitude['data'][0],
                                                     47.97,
                                                     (radar.
                                                      longitude['data'][0]),
                                                     -122.20)
    test_dis = abs(np.around(distance / 1000., 2) - 33.27)
    assert (test_dis < 0.001)


def test_get_field_location():
    # test to make sure column above location is pulled correctly
    column = pyart.util.columnsect.get_field_location(radar, 47.97, -122.20)
    # check to make sure z-gate is pulled correctly.
    test_height = abs(column.height[0] - 367)
    assert (test_height < 0.001)

    # check to make sure reflectivity value is minimum
    test_z = abs(column.reflectivity[0] + 32)
    assert (test_z < 0.001)



"""
Classes for creating and cleaning temporary directories in unit tests.

This module is taken from the nibable project.  The following license applies:

::

  The MIT License

  Copyright (c) 2009-2014 Matthew Brett <matthew.brett@gmail.com>
  Copyright (c) 2010-2013 Stephan Gerhard <git@unidesign.ch>
  Copyright (c) 2006-2014 Michael Hanke <michael.hanke@gmail.com>
  Copyright (c) 2011 Christian Haselgrove <christian.haselgrove@umassmed.edu>
  Copyright (c) 2010-2011 Jarrod Millman <jarrod.millman@gmail.com>
  Copyright (c) 2011-2014 Yaroslav Halchenko <debian@onerussian.com>

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
  THE SOFTWARE.

"""

# emacs: -*- mode: python-mode; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the NiBabel package for the
#   copyright and license terms.
#
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
""" Contexts for *with* statement providing temporary directories. """

import os
import shutil
from tempfile import template, mkdtemp


class TemporaryDirectory(object):
    """ Create and return a temporary directory. This has the same
    behavior as mkdtemp but can be used as a context manager.

    Upon exiting the context, the directory and everything contained
    in it are removed.

    Examples
    --------
    >>> import os
    >>> with TemporaryDirectory() as tmpdir:
    ...     fname = os.path.join(tmpdir, 'example_file.txt')
    ...     with open(fname, 'wt') as fobj:
    ...         _ = fobj.write('a string\\n')
    >>> os.path.exists(tmpdir)
    False
    """
    def __init__(self, suffix="", prefix=template, dir=None):
        self.name = mkdtemp(suffix, prefix, dir)
        self._closed = False

    def __enter__(self):
        return self.name

    def cleanup(self):
        if not self._closed:
            shutil.rmtree(self.name)
            self._closed = True

    def __exit__(self, exc, value, tb):
        self.cleanup()
        return False


class InTemporaryDirectory(TemporaryDirectory):
    """ Create, return, and change directory to a temporary directory.

    Examples
    --------
    >>> import os
    >>> my_cwd = os.getcwd()
    >>> with InTemporaryDirectory() as tmpdir:
    ...     _ = open('test.txt', 'wt').write('some text')
    ...     assert os.path.isfile('test.txt')
    ...     assert os.path.isfile(os.path.join(tmpdir, 'test.txt'))
    >>> os.path.exists(tmpdir)
    False
    >>> os.getcwd() == my_cwd
    True
    """
    def __enter__(self):
        self._pwd = os.getcwd()
        os.chdir(self.name)
        return super(InTemporaryDirectory, self).__enter__()

    def __exit__(self, exc, value, tb):
        os.chdir(self._pwd)
        return super(InTemporaryDirectory, self).__exit__(exc, value, tb)


class InGivenDirectory(object):
    """ Change directory to given directory for duration of ``with`` block.
    Useful when you want to use `InTemporaryDirectory` for the final test, but
    you are still debugging. For example, you may want to do this in the end:

    >>> with InTemporaryDirectory() as tmpdir:
    ...     # do something complicated which might break
    ...     pass

    But indeed the complicated thing does break, and meanwhile the
    ``InTemporaryDirectory`` context manager wiped out the directory with the
    temporary files that you wanted for debugging. So, while debugging, you
    replace with something like:

    >>> with InGivenDirectory() as tmpdir: # Use working directory by default
    ...     # do something complicated which might break
    ...     pass

    You can then look at the temporary file outputs to debug what is happening,
    fix, and finally replace ``InGivenDirectory`` with ``InTemporaryDirectory``
    again.

    """
    def __init__(self, path=None):
        """ Initialize directory context manager

        Parameters
        ----------
        path : None or str, optional
            path to change directory to, for duration of ``with`` block.
            Defaults to ``os.getcwd()`` if None.

        """
        if path is None:
            path = os.getcwd()
        self.path = os.path.abspath(path)

    def __enter__(self):
        self._pwd = os.path.abspath(os.getcwd())
        if not os.path.isdir(self.path):
            os.mkdir(self.path)
        os.chdir(self.path)
        return self.path

    def __exit__(self, exc, value, tb):
        os.chdir(self._pwd)



"""
Simple moment calculations.

"""

import numpy as np

from scipy import ndimage
from ..config import get_metadata, get_field_name
from ..core.transforms import antenna_to_cartesian
from ..util import angular_texture_2d


def calculate_snr_from_reflectivity(
        radar, refl_field=None, snr_field=None, toa=25000.):
    """
    Calculate the signal to noise ratio, in dB, from the reflectivity field.

    Parameters
    ----------
    radar : Radar
        Radar object from which to retrieve reflectivity field.
    refl_field : str, optional
        Name of field in radar which contains the reflectivity.
        None will use the default field name in the Py-ART configuration file.
    snr_field : str, optional
        Name to use for snr metadata. None will use the default field name
        in the Py-ART configuration file.
    toa : float, optional
        Height above which to take noise floor measurements, in meters.

    Returns
    -------
    snr : field dictionary
        Field dictionary containing the signal to noise ratio.

    """
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')

    range_grid = np.meshgrid(radar.range['data'],
                             np.ma.ones(radar.time['data'].shape))[0] + 1.0

    # remove range scale.. This is basically the radar constant scaled dBm
    pseudo_power = (radar.fields[refl_field]['data'] -
                    20.0*np.log10(range_grid / 1000.0))

    # Noise floor estimate
    # 25km.. should be no scatterers, not even planes, this high
    # we could get undone by AP though.. also sun
    rg, azg = np.meshgrid(radar.range['data'], radar.azimuth['data'])
    rg, eleg = np.meshgrid(radar.range['data'], radar.elevation['data'])
    _, _, z = antenna_to_cartesian(rg / 1000.0, azg, eleg)  # XXX: need to fix

    points_above = np.where(z > toa)
    noise_floor_estimate = pseudo_power[points_above].mean()

    snr_dict = get_metadata(snr_field)
    snr_dict['data'] = pseudo_power - noise_floor_estimate
    return snr_dict


def compute_noisedBZ(nrays, noisedBZ_val, _range, ref_dist,
                     noise_field=None):
    """
    Computes noise in dBZ from reference noise value.

    Parameters
    ----------
    nrays : int
        Number of rays in the reflectivity field.
    noisedBZ_val : float
        Estimated noise value in dBZ at reference distance.
    _range : np array of floats
        Range vector in m.
    ref_dist : float
        Reference distance in Km.
    noise_field : str, optional
        Name of the noise field.

    Returns
    -------
    noisedBZ : dict
        The noise field.

    """
    # parse the field parameters
    if noise_field is None:
        noise_field = get_field_name('noisedBZ_hh')

    noisedBZ_vec = noisedBZ_val+20.*np.ma.log10(1e-3*_range/ref_dist)

    noisedBZ = get_metadata(noise_field)
    noisedBZ['data'] = np.tile(noisedBZ_vec, (nrays, 1))

    return noisedBZ


def compute_snr(radar, refl_field=None, noise_field=None, snr_field=None):
    """
    Computes SNR from a reflectivity field and the noise in dBZ.

    Parameters
    ----------
    radar : Radar
        Radar object
    refl_field : str, optional
        Name of the reflectivity field to use.
    noise_field : str, optional
        Name of the noise field to use.
    snr_field : str, optional
        Name of the SNR field.

    Returns
    -------
    snr : dict
        The SNR field.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if noise_field is None:
        noise_field = get_field_name('noisedBZ_hh')
    if snr_field is None:
        snr_field = get_field_name('signal_to_noise_ratio')

    # extract fields from radar
    if refl_field in radar.fields:
        refl = radar.fields[refl_field]['data']
    else:
        raise KeyError('Field not available: ' + refl_field)
    if noise_field in radar.fields:
        noisedBZ = radar.fields[noise_field]['data']
    else:
        raise KeyError('Field not available: ' + noise_field)

    snr_data = refl-noisedBZ

    snr = get_metadata(snr_field)
    snr['data'] = snr_data

    return snr


def compute_l(radar, rhohv_field=None, l_field=None):
    """
    Computes Rhohv in logarithmic scale according to L=-log10(1-RhoHV).

    Parameters
    ----------
    radar : Radar
        Radar object.
    rhohv_field : str, optional
        Name of the RhoHV field to use.
    l_field : str, optional
        Name of the L field.

    Returns
    -------
    l : dict
        L field.

    """
    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if l_field is None:
        l_field = get_field_name('logarithmic_cross_correlation_ratio')

    # extract rhohv field from radar
    if rhohv_field in radar.fields:
        rhohv = radar.fields[rhohv_field]['data']
    else:
        raise KeyError('Field not available: ' + rhohv_field)

    rhohv[rhohv >= 1.] = 0.9999
    l_data = -np.ma.log10(1.-rhohv)

    l = get_metadata(l_field)
    l['data'] = l_data

    return l


def compute_cdr(radar, rhohv_field=None, zdr_field=None, cdr_field=None):
    """
    Computes the Circular Depolarization Ratio.

    Parameters
    ----------
    radar : Radar
        Radar object.
    rhohv_field : str, optional
        Name of the RhoHV field.
    zdr_field : str, optional
        Name of the ZDR field.
    cdr_field : str, optional
        Name of the CDR field.

    Returns
    -------
    cdr : dict
        CDR field.

    """
    # parse the field parameters
    if rhohv_field is None:
        rhohv_field = get_field_name('cross_correlation_ratio')
    if zdr_field is None:
        zdr_field = get_field_name('differential_reflectivity')
    if cdr_field is None:
        cdr_field = get_field_name('circular_depolarization_ratio')

    # extract fields from radar
    if rhohv_field in radar.fields:
        rhohv = radar.fields[rhohv_field]['data']
    else:
        raise KeyError('Field not available: ' + rhohv_field)
    if zdr_field in radar.fields:
        zdrdB = radar.fields[zdr_field]['data']
    else:
        raise KeyError('Field not available: ' + zdr_field)

    zdr = np.ma.power(10., 0.1*zdrdB)

    cdr_data = (
        10.*np.ma.log10(
            (1.+1./zdr-2.*rhohv*np.ma.sqrt(1./zdr)) /
            (1.+1./zdr+2.*rhohv*np.ma.sqrt(1./zdr))))

    cdr = get_metadata(cdr_field)
    cdr['data'] = cdr_data

    return cdr


def calculate_velocity_texture(radar, vel_field=None, wind_size=4, nyq=None,
                               check_nyq_uniform=True):
    """
    Derive the texture of the velocity field.

    Parameters
    ----------
    radar: Radar
        Radar object from which velocity texture field will be made.
    vel_field : str, optional
        Name of the velocity field. A value of None will force Py-ART to
        automatically determine the name of the velocity field.
    wind_size : int, optional
        The size of the window to calculate texture from. The window is
        defined to be a square of size wind_size by wind_size.
    nyq : float, optional
        The nyquist velocity of the radar. A value of None will force Py-ART
        to try and determine this automatically.
    check_nyquist_uniform : bool, optional
        True to check if the Nyquist velocities are uniform for all rays
        within a sweep, False will skip this check. This parameter is ignored
        when the nyq parameter is not None.

    Returns
    -------
    vel_dict: dict
        A dictionary containing the field entries for the radial velocity
        texture.

    """
    # Parse names of velocity field
    if vel_field is None:
        vel_field = get_field_name('velocity')

    # Allocate memory for texture field
    vel_texture = np.zeros(radar.fields[vel_field]['data'].shape)

    # If an array of nyquist velocities is derived, use different
    # nyquist velocites for each sweep in texture calculation according to
    # the nyquist velocity in each sweep.

    if nyq is None:
        # Find nyquist velocity if not specified
        nyq = [radar.get_nyquist_vel(i, check_nyq_uniform) for i in
               range(radar.nsweeps)]
        for i in range(0, radar.nsweeps):
            start_ray, end_ray = radar.get_start_end(i)
            inds = range(start_ray, end_ray)
            vel_sweep = radar.fields[vel_field]['data'][inds]
            vel_texture[inds] = angular_texture_2d(
                vel_sweep, wind_size, nyq[i])
    else:
        vel_texture = angular_texture_2d(
            radar.fields[vel_field]['data'], wind_size, nyq)
    vel_texture_field = get_metadata('velocity')
    vel_texture_field['long_name'] = 'Doppler velocity texture'
    vel_texture_field['standard_name'] = (
        'texture_of_radial_velocity' + '_of_scatters_away_from_instrument')
    vel_texture_field['data'] = ndimage.filters.median_filter(vel_texture,
                                                              size=(wind_size,
                                                                    wind_size))
    return vel_texture_field



""" Unit Tests for Py-ART's core/radar.py module. """

import pickle
import sys
# we need a class which excepts str for writing in Python 2 and 3
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import inspect

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
import pytest

import pyart
from pyart.lazydict import LazyLoadDict


def test_radar_picklable():
    # verify that Radar instances are picklable
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    picklestring = pickle.dumps(radar)
    radar_new = pickle.loads(picklestring)
    assert 'data' in radar.gate_x
    assert 'data' in radar_new.gate_x


def test_gate_longitude_latitude():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.azimuth['data'][:] = [0, 90, 180, 270, 0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0, 10, 10, 10, 10]
    radar.range['data'][:] = [5, 15, 25, 35, 45]

    assert radar.gate_longitude['data'].shape == (8, 5)
    assert radar.gate_latitude['data'].shape == (8, 5)
    assert_almost_equal(radar.gate_longitude['data'][0, 0], -97.5, 1)
    assert_almost_equal(radar.gate_latitude['data'][0, 0], 36.5, 1)

    # reset and try again with a non-default lat_0/lon_0
    radar.init_gate_longitude_latitude()
    radar.projection.pop('_include_lon_0_lat_0')
    radar.projection['lat_0'] = 20.0
    radar.projection['lon_0'] = 60.0

    assert_almost_equal(radar.gate_longitude['data'][0, 0], 60.0, 1)
    assert_almost_equal(radar.gate_latitude['data'][0, 0], 20.0, 1)


def test_gate_altitude():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.azimuth['data'][:] = [0, 90, 180, 270, 0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0, 10, 10, 10, 10]
    radar.range['data'][:] = [5, 15, 25, 35, 45]

    assert radar.gate_altitude['data'].shape == (8, 5)
    assert_almost_equal(radar.gate_altitude['data'][0, 0], 200.0, 1)

    radar.init_gate_altitude()
    radar.altitude['data'][0] = 150.
    assert_almost_equal(radar.gate_altitude['data'][0, 0], 150.0, 1)


def test_gate_x_y_z():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.azimuth['data'][:] = [0, 90, 180, 270, 0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0, 10, 10, 10, 10]
    radar.range['data'][:] = [5, 15, 25, 35, 45]

    assert radar.gate_x['data'].shape == (8, 5)
    assert_allclose(radar.gate_x['data'][0], [0, 0, 0, 0, 0], atol=1e-14)
    assert_allclose(radar.gate_x['data'][1], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(radar.gate_x['data'][2], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(
        radar.gate_x['data'][3], [-5, -15, -25, -35, -45], atol=1e-14)

    assert radar.gate_y['data'].shape == (8, 5)
    assert_allclose(radar.gate_y['data'][0], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(radar.gate_y['data'][1], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(
        radar.gate_y['data'][2], [-5, -15, -25, -35, -45], atol=1e-14)
    assert_allclose(radar.gate_y['data'][3], [0, 0, 0, 0, 0], atol=1e-6)

    assert radar.gate_z['data'].shape == (8, 5)
    z_sweep0 = np.array([1.47e-6, 1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4])
    assert_allclose(radar.gate_z['data'][0], z_sweep0, atol=1e-3)
    assert_allclose(radar.gate_z['data'][1], z_sweep0, atol=1e-3)
    assert_allclose(radar.gate_z['data'][2], z_sweep0, atol=1e-3)
    assert_allclose(radar.gate_z['data'][3], z_sweep0, atol=1e-3)


def test_get_gate_x_y_z():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.azimuth['data'][:] = [0, 90, 180, 270, 0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0, 10, 10, 10, 10]
    radar.range['data'][:] = [5, 15, 25, 35, 45]

    gate_x, gate_y, gate_z = radar.get_gate_x_y_z(0)
    assert gate_x.shape == (4, 5)
    assert_allclose(gate_x[0], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(gate_x[1], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(gate_x[2], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(gate_x[3], [-5, -15, -25, -35, -45], atol=1e-14)

    assert gate_y.shape == (4, 5)
    assert_allclose(gate_y[0], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(gate_y[1], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(gate_y[2], [-5, -15, -25, -35, -45], atol=1e-14)
    assert_allclose(gate_y[3], [0, 0, 0, 0, 0], atol=1e-5)

    assert gate_z.shape == (4, 5)
    z_sweep0 = np.array([1.47e-6, 1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4])
    assert_allclose(gate_z[0], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[1], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[2], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[3], z_sweep0, atol=1e-3)


def test_get_gate_x_y_z_edges():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 1)
    radar.azimuth['data'][:] = [315., 45, 135., 225.]
    radar.elevation['data'][:] = [0, 0, 0, 0]
    radar.range['data'][:] = [5, 15, 25, 35, 45]

    gate_x, gate_y, gate_z = radar.get_gate_x_y_z(0, edges=True)

    zeros = np.array([0, 0, 0, 0, 0, 0])
    even = np.array([0, 10, 20, 30, 40, 50])
    atol = 1e-4

    assert gate_x.shape == (5, 6)
    assert_allclose(gate_x[0], -even, atol=atol)
    assert_allclose(gate_x[1], zeros, atol=atol)
    assert_allclose(gate_x[2], even, atol=atol)
    assert_allclose(gate_x[3], zeros, atol=atol)
    assert_allclose(gate_x[4], -even, atol=atol)

    assert gate_y.shape == (5, 6)
    assert_allclose(gate_y[0], zeros, atol=atol)
    assert_allclose(gate_y[1], even, atol=atol)
    assert_allclose(gate_y[2], zeros, atol=atol)
    assert_allclose(gate_y[3], -even, atol=atol)
    assert_allclose(gate_y[4], zeros, atol=atol)

    assert gate_z.shape == (5, 6)
    z_sweep0 = np.array([0, 5.89e-6, 2.354e-5, 5.297e-5, 9.418e-5, 1.4715e-4])
    assert_allclose(gate_z[0], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[1], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[2], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[3], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[4], z_sweep0, atol=1e-3)


def test_get_gate_x_y_z_transitions():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.azimuth['data'][:] = [0, 90, 180, 270, 0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0, 10, 10, 10, 10]
    radar.range['data'][:] = [5, 15, 25, 35, 45]
    radar.antenna_transition = {'data': np.array([0, 0, 1, 0, 0, 0, 0, 0])}

    gate_x, gate_y, gate_z = radar.get_gate_x_y_z(0, filter_transitions=True)
    assert gate_x.shape == (3, 5)
    assert_allclose(gate_x[0], [0, 0, 0, 0, 0], atol=1e-14)
    assert_allclose(gate_x[1], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(gate_x[2], [-5, -15, -25, -35, -45], atol=1e-14)

    assert gate_y.shape == (3, 5)
    assert_allclose(gate_y[0], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(gate_y[1], [0, 0, 0, 0, 0], atol=1e-5)
    assert_allclose(gate_y[2], [0, 0, 0, 0, 0], atol=1e-5)

    assert gate_z.shape == (3, 5)
    z_sweep0 = np.array([1.47e-6, 1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4])
    assert_allclose(gate_z[0], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[1], z_sweep0, atol=1e-3)
    assert_allclose(gate_z[2], z_sweep0, atol=1e-3)


def test_get_gate_lat_lon_alt():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    lat, lon, alt = radar.get_gate_lat_lon_alt(0)
    assert lat.shape == (4, 5)
    assert_allclose(lat[0], [36.5, 36.502243, 36.50449, 36.506744, 36.50899],
                    atol=1e-3)
    assert_allclose(lat[1], [36.5, 36.502243, 36.50449, 36.506737, 36.508984],
                    atol=1e-3)
    assert_allclose(lat[2], [36.5, 36.502243, 36.50449, 36.506737, 36.508984],
                    atol=1e-3)
    assert_allclose(lat[3], [36.5, 36.50224, 36.504486, 36.50673 , 36.508976],
                    atol=1e-3)

    assert lon.shape == (4, 5)
    assert_allclose(lon[0], [-97.49999, -97.49999, -97.49999, -97.49999, -97.49999],
                    atol=1e-3)
    assert_allclose(lon[1], [-97.49999, -97.49995, -97.4999, -97.499855, -97.499794],
                    atol=1e-3)
    assert_allclose(lon[2], [-97.49999, -97.4999, -97.499794, -97.4997, -97.4996],
                    atol=1e-3)
    assert_allclose(lon[3], [-97.49999, -97.499855, -97.4997, -97.49956, -97.499405],
                    atol=1e-3)

    assert lat.shape == (4, 5)
    alt_sweep0 = np.array([200., 203., 206., 209., 213.])
    assert_allclose(alt[0], alt_sweep0, atol=1e-3)
    assert_allclose(alt[1], alt_sweep0, atol=1e-3)
    assert_allclose(alt[2], alt_sweep0, atol=1e-3)
    assert_allclose(alt[3], alt_sweep0, atol=1e-3)


def test_get_gate_lat_lon_alt_transitions():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.antenna_transition = {'data': np.array([0, 0, 1, 0, 0, 0, 0, 0])}
    lat, lon, alt = radar.get_gate_lat_lon_alt(0, filter_transitions=True)
    assert lat.shape == (3, 5)
    assert_allclose(lat[0], [36.5, 36.502243, 36.50449, 36.506744, 36.50899],
                    atol=1e-3)
    assert_allclose(lat[1], [36.5, 36.502243, 36.50449, 36.506737, 36.508984],
                    atol=1e-3)
    assert_allclose(lat[2], [36.5, 36.50224, 36.504486, 36.50673, 36.508976],
                    atol=1e-3)

    assert lon.shape == (3, 5)
    assert_allclose(lon[0], [-97.49999, -97.49999, -97.49999, -97.49999, -97.49999],
                    atol=1e-3)
    assert_allclose(lon[1], [-97.49999, -97.49995, -97.4999, -97.499855, -97.499794],
                    atol=1e-3)
    assert_allclose(lon[2], [-97.49999, -97.499855, -97.4997, -97.49956, -97.499405],
                    atol=1e-3)

    assert lat.shape == (3, 5)
    alt_sweep0 = np.array([200., 203., 206., 209., 213.])
    assert_allclose(alt[0], alt_sweep0, atol=1e-3)
    assert_allclose(alt[1], alt_sweep0, atol=1e-3)
    assert_allclose(alt[2], alt_sweep0, atol=1e-3)


def test_init_gate_x_y_z():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 1)
    radar.azimuth['data'][:] = [0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0]
    radar.range['data'][:] = [5, 15, 25, 35, 45]

    # access and check initial gate locations
    assert_allclose(radar.gate_x['data'][1], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(radar.gate_y['data'][0], [5, 15, 25, 35, 45], atol=1e-14)
    z_sweep0 = np.array([1.47e-6, 1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4])
    assert_allclose(radar.gate_z['data'][0], z_sweep0, atol=1e-3)

    # change range, gate_x, y, z are not updated
    radar.range['data'][:] = [15, 25, 35, 45, 55]
    assert_allclose(radar.gate_x['data'][1], [5, 15, 25, 35, 45], atol=1e-14)
    assert_allclose(radar.gate_y['data'][0], [5, 15, 25, 35, 45], atol=1e-14)
    z_sweep0 = np.array([1.47e-6, 1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4])
    assert_allclose(radar.gate_z['data'][0], z_sweep0, atol=1e-3)

    # call init_gate_x_y_z, now the attributes are updated
    radar.init_gate_x_y_z()
    assert_allclose(radar.gate_x['data'][1], [15, 25, 35, 45, 55], atol=1e-14)
    assert_allclose(radar.gate_y['data'][0], [15, 25, 35, 45, 55], atol=1e-14)
    z_sweep0 = np.array([1.324e-5, 3.679e-5, 7.210e-5, 1.1919e-4, 1.7805e-4])
    assert_allclose(radar.gate_z['data'][0], z_sweep0, atol=1e-3)


def test_rays_per_sweep_attribute():
    radar = pyart.testing.make_target_radar()
    rays_per_sweep = radar.rays_per_sweep
    assert isinstance(rays_per_sweep, LazyLoadDict)
    assert rays_per_sweep['data'].shape == (1, )
    assert rays_per_sweep['data'][0] == 360


def test_iterators():
    radar = pyart.testing.make_empty_ppi_radar(30, 20, 5)
    radar.fields['reflectivity'] = {
        'data': np.zeros((100, 30), dtype=np.float32)}

    starts = [0, 20, 40, 60, 80]
    ends = [19, 39, 59, 79, 99]
    starts_ends = [(s, e) for s, e in zip(starts, ends)]

    assert inspect.isgenerator(radar.iter_start())
    assert [s for s in radar.iter_start()] == starts

    assert inspect.isgenerator(radar.iter_end())
    assert [s for s in radar.iter_end()] == ends

    assert inspect.isgenerator(radar.iter_start_end())
    assert [s for s in radar.iter_start_end()] == starts_ends

    assert inspect.isgenerator(radar.iter_slice())
    for s, start, end in zip(radar.iter_slice(), starts, ends):
        assert s.start == start
        assert s.stop == end + 1
        assert s.step is None

    assert inspect.isgenerator(radar.iter_field('reflectivity'))
    for d in radar.iter_field('reflectivity'):
        assert d.shape == (20, 30)
        assert d.dtype == np.float32
    pytest.raises(KeyError, radar.iter_field, 'foobar')

    assert inspect.isgenerator(radar.iter_azimuth())
    for d in radar.iter_azimuth():
        assert d.shape == (20, )

    assert inspect.isgenerator(radar.iter_elevation())
    for d in radar.iter_elevation():
        assert d.shape == (20, )


def test_get_methods():
    radar = pyart.testing.make_empty_ppi_radar(30, 20, 5)
    radar.fields['reflectivity'] = {
        'data': np.zeros((100, 30), dtype=np.float32)}

    assert radar.get_start(0) == 0
    assert radar.get_start(1) == 20
    pytest.raises(IndexError, radar.get_start, -1)
    pytest.raises(IndexError, radar.get_start, 20)

    assert radar.get_end(0) == 19
    assert radar.get_end(1) == 39
    pytest.raises(IndexError, radar.get_end, -1)
    pytest.raises(IndexError, radar.get_end, 20)

    assert radar.get_start_end(0) == (0, 19)
    assert radar.get_start_end(1) == (20, 39)
    pytest.raises(IndexError, radar.get_start_end, -1)
    pytest.raises(IndexError, radar.get_start_end, 20)

    assert radar.get_slice(0) == slice(0, 20)
    assert radar.get_slice(1) == slice(20, 40)
    pytest.raises(IndexError, radar.get_slice, -1)
    pytest.raises(IndexError, radar.get_slice, 20)

    data = radar.get_field(0, 'reflectivity')
    assert data.shape == (20, 30)
    assert data.dtype == np.float32
    data = radar.get_field(1, 'reflectivity')
    assert data.shape == (20, 30)
    assert data.dtype == np.float32
    pytest.raises(KeyError, radar.get_field, 0, 'foobar')
    pytest.raises(IndexError, radar.get_field, -1, 'reflectivity')
    pytest.raises(IndexError, radar.get_field, 20, 'reflectivity')

    assert radar.get_azimuth(0).shape == (20, )
    pytest.raises(IndexError, radar.get_azimuth, -1)
    pytest.raises(IndexError, radar.get_azimuth, 20)

    assert radar.get_elevation(0).shape == (20, )
    pytest.raises(IndexError, radar.get_elevation, -1)
    pytest.raises(IndexError, radar.get_elevation, 20)

    pytest.raises(LookupError, radar.get_nyquist_vel, 0)
    radar.instrument_parameters = {
        'nyquist_velocity': {'data': np.ones((100,))}
    }
    assert round(radar.get_nyquist_vel(0)) == 1
    pytest.raises(IndexError, radar.get_nyquist_vel, -1)
    radar.instrument_parameters['nyquist_velocity']['data'][0] = 2
    pytest.raises(Exception, radar.get_nyquist_vel, 0)


def test_get_gate_area():
    radar = pyart.testing.make_empty_ppi_radar(5, 4, 2)
    radar.azimuth['data'][:] = [0, 90, 180, 270, 0, 90, 180, 270]
    radar.elevation['data'][:] = [0, 0, 0, 0, 10, 10, 10, 10]
    radar.range['data'][:] = [5, 15, 25, 35, 45]
    area = radar.get_gate_area(0)
    assert_allclose(
        area[0], [157.07964, 314.1593, 471.23892, 628.3185], atol=1e-3)


def test_extract_sweeps():
    radar = pyart.testing.make_empty_ppi_radar(100, 360, 3)
    radar.fields['reflectivity'] = {'data': np.zeros((1080, 100))}
    radar.fields['velocity'] = {'data': np.zeros((1080, 100))}

    eradar = radar.extract_sweeps([0, 2])

    # extracted radar should have 720 rays, 2 sweeps, 100 gates
    assert eradar.time['data'].shape == (720, )
    assert eradar.range['data'].shape == (100, )

    assert eradar.metadata['instrument_name'] == 'fake_radar'
    assert eradar.scan_type == 'ppi'

    assert eradar.latitude['data'].shape == (1, )
    assert eradar.longitude['data'].shape == (1, )
    assert eradar.altitude['data'].shape == (1, )
    assert eradar.altitude_agl is None

    assert eradar.sweep_number['data'].shape == (2, )
    assert eradar.sweep_mode['data'].shape == (2, )
    assert eradar.fixed_angle['data'].shape == (2, )
    assert eradar.sweep_start_ray_index['data'].shape == (2, )
    assert eradar.sweep_end_ray_index['data'].shape == (2, )
    assert eradar.target_scan_rate is None

    assert eradar.azimuth['data'].shape == (720, )
    assert eradar.elevation['data'].shape == (720, )
    assert eradar.scan_rate is None
    assert eradar.antenna_transition is None

    assert eradar.instrument_parameters is None
    assert eradar.radar_calibration is None

    assert eradar.ngates == 100
    assert eradar.nrays == 720
    assert eradar.nsweeps == 2

    assert eradar.fields['reflectivity']['data'].shape == (720, 100)
    assert eradar.fields['velocity']['data'].shape == (720, 100)


def test_extract_sweeps_extra():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 3)
    radar.instrument_parameters = {
        'prt': {'data': np.zeros((108, ))},
        'prt_mode': {'data': np.array(['fixed'] * 3)},
        'radar_antenna_gain_h': {'data': np.array(0)},
    }

    radar.radar_calibration = {
        'r_calib_index': {'data': np.zeros((108, ))},
        'r_calib_time': {'data': np.zeros((8, ))}
    }

    eradar = radar.extract_sweeps([0, 2])

    instr = eradar.instrument_parameters
    assert instr['prt']['data'].shape == (72, )
    assert instr['prt_mode']['data'].shape == (2, )
    assert instr['radar_antenna_gain_h']['data'].shape == ()

    calib = eradar.radar_calibration
    assert calib['r_calib_index']['data'].shape == (72, )
    assert calib['r_calib_time']['data'].shape == (8, )


def test_extract_sweeps_errors():
    radar = pyart.testing.make_empty_ppi_radar(10, 36, 2)
    pytest.raises(ValueError, radar.extract_sweeps, [0, 2])
    pytest.raises(ValueError, radar.extract_sweeps, [-1, 1])


def test_radar_creation():
    radar = pyart.testing.make_target_radar()
    assert isinstance(radar, pyart.core.Radar)


def test_add_field():
    radar = pyart.testing.make_target_radar()
    dic = {'data': np.zeros((360, 50)), 'standard_name': 'test'}
    radar.add_field('test', dic)
    assert 'test' in radar.fields
    assert 'data' in radar.fields['test']
    assert radar.fields['test']['standard_name'] == 'test'


def test_add_field_errors():
    radar = pyart.testing.make_target_radar()

    pytest.raises(ValueError, radar.add_field, 'reflectivity', {})

    dic = {'dat': np.zeros((360, 50)), 'standard_name': 'test'}
    pytest.raises(KeyError, radar.add_field, 'test', dic)

    dic = {'data': np.zeros((360, 49)), 'standard_name': 'test'}
    pytest.raises(ValueError, radar.add_field, 'test', dic)


def test_add_field_like():
    radar = pyart.testing.make_target_radar()
    data = np.zeros((360, 50))
    radar.add_field_like('reflectivity', 'test', data)
    assert 'test' in radar.fields
    assert 'data' in radar.fields['test']
    assert radar.fields['test']['units'] == 'dBZ'


def test_add_field_like_bug():
    # tests for bug where adding a field over-writes 'like' field
    # data/metadata.
    radar = pyart.testing.make_target_radar()
    data = np.ones((360, 50))
    radar.add_field_like('reflectivity', 'test', data)
    radar.fields['test']['units'] = 'fake'

    # check field added
    assert radar.fields['test']['units'] == 'fake'
    assert radar.fields['test']['data'][0, 0] == 1

    # check original field
    assert radar.fields['reflectivity']['units'] == 'dBZ'
    assert radar.fields['reflectivity']['data'][0, 0] == 0


def test_add_field_like_errors():
    radar = pyart.testing.make_target_radar()
    pytest.raises(ValueError, radar.add_field_like, 'foo', 'bar', [])


@pytest.mark.parametrize(
    "level", ['standard', 's', 'compact', 'c', 'full', 'f'])
def test_info_levels(level):
    check_info(level)


def test_info_nonstandard():
    # build a non-standard radar object for testing all paths in info
    radar = pyart.testing.make_target_radar()
    radar.fields['reflectivity']['data'] = [1, 2, 3, 4]
    radar.instrument_parameters = {'foobar': {'data': [1, 2], 'bar': 'foo'}}
    radar.radar_calibration = {'foobar': {'data': [1, 2], 'bar': 'foo'}}
    check_info('standard', radar)


def check_info(level, radar=None):
    out = StringIO()
    get_info(level, out, radar)
    # don't check the output, just that something was printed.
    assert len(out.getvalue()) != 0


def get_info(level='standard', out=sys.stdout, radar=None):
    if radar is None:
        radar = pyart.testing.make_target_radar()
    radar.info(level, out)


def test_info_errors():
    pytest.raises(ValueError, check_info, 'foo')



"""
Functions for rainfall rate estimation.

"""

from warnings import warn

import numpy as np

from ..config import get_metadata, get_field_name, get_fillvalue
from .echo_class import get_freq_band


def est_rain_rate_zpoly(radar, refl_field=None, rr_field=None):
    """
    Estimates rainfall rate from reflectivity using a polynomial Z-R relation
    developed at McGill University.

    Parameters
    ----------
    radar : Radar
        Radar object.
    refl_field : str, optional
        Name of the reflectivity field to use.
    rr_field : str, optional
        Name of the rainfall rate field.

    Returns
    -------
    rain : dict
        Field dictionary containing the rainfall rate.

    References
    ----------
    Doelling et al. Systematic variations of Z–R-relationships from drop size
    distributions measured in northern Germany during seven years. 1998. Atmos.
    Ocean. Technol, 21, 1545-1556.

    Joss et al. Operational Use of Radar for Precipitation Measurements
    in Switzerland. 1998. Vdf Hochschulverlag AG ETH Zurich: 134.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    refl2 = refl*refl
    refl3 = refl*refl2
    refl4 = refl*refl3

    rr_data = np.ma.power(
        10., -2.3+0.17*refl-5.1e-3*refl2+9.8e-5*refl3-6e-7*refl4)

    rain = get_metadata(rr_field)
    rain['data'] = rr_data

    return rain


def est_rain_rate_z(radar, alpha=0.0376, beta=0.6112, refl_field=None,
                    rr_field=None):
    """
    Estimates rainfall rate from reflectivity using a power law.

    Parameters
    ----------
    radar : Radar
        Radar object.
    alpha, beta : floats, optional
        Factor (alpha) and exponent (beta) of the power law.
    refl_field : str, optional
        Name of the reflectivity field to use.
    rr_field : str, optional
        Name of the rainfall rate field.

    Returns
    -------
    rain : dict
        Field dictionary containing the rainfall rate.

    Reference
    ---------
    Fabry, Frédéric. Radar Meterology. 2015. Ch 9. pg 148-165.
    https://doi.org/10.1017/CBO9781107707405

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    radar.check_field_exists(refl_field)
    refl = radar.fields[refl_field]['data']

    rr_data = alpha*np.ma.power(np.ma.power(10., 0.1*refl), beta)

    rain = get_metadata(rr_field)
    rain['data'] = rr_data

    return rain


def est_rain_rate_kdp(radar, alpha=None, beta=None, kdp_field=None,
                      rr_field=None):
    """
    Estimates rainfall rate from kdp using alpha power law.

    Parameters
    ----------
    radar : Radar
        Radar object.
    alpha, beta : floats, optional
        Factor (alpha) and exponent (beta) of the power law. If not set the
        factors are going to be determined according to the radar frequency.
    kdp_field : str, optional
        Name of the specific differential phase field to use.
    rr_field : str, optional
        Name of the rainfall rate field.

    Returns
    -------
    rain : dict
        Field dictionary containing the rainfall rate.

    Reference
    ---------
    Figueras et al. Long-term monitoring of French polarimetric radar data
    quality and evaluation of several polarimetric quantitative precipitation
    estimators in ideal conditions for operational implementation at C-band.
    Quarterly Journal of the Royal Meteorological Society. 2012.
    https://doi.org/10.1002/qj.1934

    """
    # select the coefficients as alpha function of frequency band
    if alpha is None or beta is None:
        # assign coefficients according to radar frequency
        if 'frequency' in radar.instrument_parameters:
            alpha, beta = _get_coeff_rkdp(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            alpha, beta = _coeff_rkdp_table()['C']
            warn('Radar frequency unknown. '
                 + 'Default coefficients for C band will be applied.')

    # parse the field parameters
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    radar.check_field_exists(kdp_field)
    kdp = radar.fields[kdp_field]['data']

    kdp[kdp < 0] = 0.
    rr_data = alpha*np.ma.power(kdp, beta)

    rain = get_metadata(rr_field)
    rain['data'] = rr_data

    return rain


def est_rain_rate_a(radar, alpha=None, beta=None, a_field=None,
                    rr_field=None):
    """
    Estimates rainfall rate from specific attenuation using alpha power law.

    Parameters
    ----------
    radar : Radar
        Radar object.
    alpha, beta : floats, optional
        Factor (alpha) and exponent (beta) of the power law. If not set the
        factors are going to be determined according to the radar frequency.
    a_field : str, optional
        Name of the specific attenuation field to use.
    rr_field : str, optional
        Name of the rainfall rate field.

    Returns
    -------
    rain : dict
        Field dictionary containing the rainfall rate.

    References
    ----------
    Diederich M., Ryzhkov A., Simmer C., Zhang P. and Tromel S., 2015: Use of
    Specific Attenuation for Rainfall Measurement at X-Band Radar Wavelenghts.
    Part I: Radar Calibration and Partial Beam Blockage Estimation. Journal of
    Hydrometeorology, 16, 487-502.

    Ryzhkov A., Diederich M., Zhang P. and Simmer C., 2014: Potential
    Utilization of Specific Attenuation for Rainfall Estimation, Mitigation of
    Partial Beam Blockage, and Radar Networking. Journal of Atmospheric and
    Oceanic Technology, 31, 599-619.

    """
    # select the coefficients as alpha function of frequency band
    if alpha is None or beta is None:
        # assign coefficients according to radar frequency
        if 'frequency' in radar.instrument_parameters:
            alpha, beta = _get_coeff_ra(
                radar.instrument_parameters['frequency']['data'][0])
        else:
            alpha, beta = _coeff_ra_table()['C']
            warn('Radar frequency unknown. '
                 + 'Default coefficients for C band will be applied.')

    # parse the field parameters
    if a_field is None:
        a_field = get_field_name('specific_attenuation')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    radar.check_field_exists(a_field)
    att = radar.fields[a_field]['data']

    rr_data = alpha*np.ma.power(att, beta)

    rain = get_metadata(rr_field)
    rain['data'] = rr_data

    return rain


def est_rain_rate_zkdp(radar, alphaz=0.0376, betaz=0.6112, alphakdp=None,
                       betakdp=None, refl_field=None, kdp_field=None,
                       rr_field=None, main_field=None, thresh=None,
                       thresh_max=True):
    """
    Estimates rainfall rate from a blending of power law r-kdp and r-z
    relations.

    Parameters
    ----------
    radar : Radar
        Radar object.
    alphaz, betaz : floats, optional
        Factor (alpha) and exponent (beta) of the z-r power law.
    alphakdp, betakdp : floats, optional
        Factor (alpha) and exponent (beta) of the kdp-r power law.
        If not set the factors are going to be determined according
        to the radar frequency.
    refl_field : str, optional
        Name of the reflectivity field to use.
    kdp_field : str, optional
        Name of the specific differential phase field to use.
    rr_field : str, optional
        Name of the rainfall rate field.
    main_field : str, optional
        Name of the field that is going to act as main. Has to be
        either refl_field or kdp_field. Default is refl_field.
    thresh : float, optional
        Value of the threshold that determines when to use the secondary
        field.
    thresh_max : Bool, optional
        If true the main field is used up to the thresh value maximum.
        Otherwise the main field is not used below thresh value.

    Returns
    -------
    rain_main : dict
        Field dictionary containing the rainfall rate.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if kdp_field is None:
        kdp_field = get_field_name('specific_differential_phase')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    rain_z = est_rain_rate_z(
        radar, alpha=alphaz, beta=betaz, refl_field=refl_field,
        rr_field=rr_field)
    rain_kdp = est_rain_rate_kdp(
        radar, alpha=alphakdp, beta=betakdp, kdp_field=kdp_field,
        rr_field=rr_field)

    if main_field == refl_field:
        secondary_field = kdp_field
        rain_main = rain_z
        rain_secondary = rain_kdp
    elif main_field == kdp_field:
        secondary_field = refl_field
        rain_main = rain_kdp
        rain_secondary = rain_z
    elif main_field is None:
        main_field = refl_field
        secondary_field = kdp_field
        rain_main = rain_z
        rain_secondary = rain_kdp
    else:
        main_field = refl_field
        secondary_field = kdp_field
        rain_main = rain_z
        rain_secondary = rain_kdp
        thresh = 40.
        thresh_max = True
        warn('Unknown main field. Using ' + refl_field
             + ' with threshold ' + str(thresh))

    if thresh_max:
        is_secondary = rain_main['data'] > thresh
    else:
        is_secondary = rain_main['data'] < thresh
    rain_main['data'][is_secondary] = (
        rain_secondary['data'][is_secondary])

    return rain_main


def est_rain_rate_za(radar, alphaz=0.0376, betaz=0.6112, alphaa=None,
                     betaa=None, refl_field=None, a_field=None, rr_field=None,
                     main_field=None, thresh=None, thresh_max=False):
    """
    Estimates rainfall rate from a blending of power law r-alpha and r-z
    relations.

    Parameters
    ----------
    radar : Radar
        Radar object
    alphaz, betaz : floats, optional
        Factor (alpha) and exponent (beta) of the z-r power law.
    alphaa,betaa : floats, optional
        Factor (alpha) and exponent (beta) of the a-r power law. If not set
        the factors are going to be determined according to the radar frequency.
    refl_field : str, optional
        Name of the reflectivity field to use.
    a_field : str, optional
        Name of the specific attenuation field to use.
    rr_field : str, optional
        Name of the rainfall rate field.
    main_field : str, optional
        Name of the field that is going to act as main. Has to be
        either refl_field or kdp_field. Default is refl_field.
    thresh : float, optional
        Value of the threshold that determines when to use the secondary
        field.
    thresh_max : Bool, optional
        If true the main field is used up to the thresh value maximum.
        Otherwise the main field is not used below thresh value.

    Returns
    -------
    rain_main : dict
        Field dictionary containing the rainfall rate.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if a_field is None:
        a_field = get_field_name('specific_attenuation')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    rain_z = est_rain_rate_z(
        radar, alpha=alphaz, beta=betaz, refl_field=refl_field,
        rr_field=rr_field)
    rain_a = est_rain_rate_a(
        radar, alpha=alphaa, beta=betaa, a_field=a_field, rr_field=rr_field)

    if main_field == refl_field:
        secondary_field = a_field
        rain_main = rain_z
        rain_secondary = rain_a
    elif main_field == a_field:
        secondary_field = refl_field
        rain_main = rain_a
        rain_secondary = rain_z
    elif main_field is None:
        main_field = a_field
        secondary_field = refl_field
        rain_main = rain_a
        rain_secondary = rain_z
    else:
        main_field = a_field
        secondary_field = refl_field
        rain_main = rain_a
        rain_secondary = rain_z
        thresh = 0.04
        thresh_max = False
        warn('Unknown main field. Using ' + a_field + ' with threshold '
             + str(thresh))

    if thresh_max:
        is_secondary = rain_main['data'] > thresh
    else:
        is_secondary = rain_main['data'] < thresh

    rain_main['data'][is_secondary] = (
        rain_secondary['data'][is_secondary])

    return rain_main


def est_rain_rate_hydro(radar, alphazr=0.0376, betazr=0.6112, alphazs=0.1,
                        betazs=0.5, alphaa=None, betaa=None, mp_factor=0.6,
                        refl_field=None, a_field=None, hydro_field=None,
                        rr_field=None, main_field=None, thresh=None,
                        thresh_max=False):
    """
    Estimates rainfall rate using different relations between R and the
    polarimetric variables depending on the hydrometeor type.

    Parameters
    ----------
    radar : Radar
        Radar object.
    alphazr, betazr : floats, optional
        Factor (alpha) and exponent (beta) of the z-r power law for rain.
    alphazs, betazs : floats, optional
        Factor (alpha) and exponent (beta) of the z-s power law for snow.
    alphaa, betaa : floats, optional
        Factor (alpha) and exponent (beta) of the a-r power law.
        If not set the factors are going to be determined according
        to the radar frequency.
    mp_factor : float, optional
        Factor applied to z-r relation in the melting layer.
    refl_field : str, optional
        Name of the reflectivity field to use.
    a_field : str, optional
        Name of the specific attenuation field to use.
    hydro_field : str, optional
        Name of the hydrometeor classification field to use.
    rr_field : str, optional
        Name of the rainfall rate field.
    main_field : str, optional
        Name of the field that is going to act as main. Has to be
        either refl_field or kdp_field. Default is refl_field.
    thresh : float, optional
        Value of the threshold that determines when to use the secondary
        field.
    thresh_max : Bool, optional
        If true the main field is used up to the thresh value maximum.
        Otherwise the main field is not used below thresh value.

    Returns
    -------
    rain : dict
        Field dictionary containing the rainfall rate.

    """
    # parse the field parameters
    if refl_field is None:
        refl_field = get_field_name('reflectivity')
    if a_field is None:
        a_field = get_field_name('specific_attenuation')
    if hydro_field is None:
        hydro_field = get_field_name('radar_echo_classification')
    if rr_field is None:
        rr_field = get_field_name('radar_estimated_rain_rate')

    # extract fields and parameters from radar
    if hydro_field in radar.fields:
        hydroclass = radar.fields[hydro_field]['data']
    else:
        raise KeyError('Field not available: ' + hydro_field)

    # get the location of each hydrometeor class
    is_ds = hydroclass == 1
    is_cr = hydroclass == 2
    is_lr = hydroclass == 3
    is_gr = hydroclass == 4
    is_rn = hydroclass == 5
    is_vi = hydroclass == 6
    is_ws = hydroclass == 7
    is_mh = hydroclass == 8
    is_ih = hydroclass == 9

    # compute z-r (in rain) z-r in snow and z-a relations
    rain_z = est_rain_rate_z(
        radar, alpha=alphazr, beta=betazr, refl_field=refl_field,
        rr_field=rr_field)
    snow_z = est_rain_rate_z(
        radar, alpha=alphazs, beta=betazs, refl_field=refl_field,
        rr_field=rr_field)
    rain_a = est_rain_rate_a(
        radar, alpha=alphaa, beta=betaa, a_field=a_field, rr_field=rr_field)

    # initialize rainfall rate field
    rr_data = np.ma.zeros(hydroclass.shape, dtype='float32')
    rr_data[:] = np.ma.masked
    rr_data.set_fill_value(get_fillvalue())

    # apply the relations for each type
    # solid phase
    rr_data[is_ds] = snow_z['data'][is_ds]
    rr_data[is_cr] = snow_z['data'][is_cr]
    rr_data[is_vi] = snow_z['data'][is_vi]
    rr_data[is_gr] = snow_z['data'][is_gr]
    rr_data[is_ih] = snow_z['data'][is_ih]

    # rain
    if main_field == refl_field:
        secondary_field = a_field
        rain_main = rain_z
        rain_secondary = rain_a
    elif main_field == a_field:
        secondary_field = refl_field
        rain_main = rain_a
        rain_secondary = rain_z
    elif main_field is None:
        main_field = a_field
        secondary_field = refl_field
        rain_main = rain_a
        rain_secondary = rain_z
    else:
        main_field = a_field
        secondary_field = refl_field
        rain_main = rain_a
        rain_secondary = rain_z
        thresh = 0.04
        thresh_max = False
        warn('Unknown main field. Using ' + a_field + ' with threshold ' +
             str(thresh))

    if thresh_max:
        is_secondary = rain_main['data'] > thresh
    else:
        is_secondary = rain_main['data'] < thresh

    rain_main['data'][is_secondary] = rain_secondary['data'][is_secondary]

    rr_data[is_lr] = rain_main['data'][is_lr]
    rr_data[is_rn] = rain_main['data'][is_rn]

    # mixed phase
    rr_data[is_ws] = mp_factor*rain_z['data'][is_ws]
    rr_data[is_mh] = mp_factor*rain_z['data'][is_mh]

    rain = get_metadata(rr_field)
    rain['data'] = rr_data

    return rain


def _get_coeff_rkdp(freq):
    """
    Get the R(kdp) power law coefficients for a particular frequency.

    Parameters
    ----------
    freq : float
        Radar frequency [Hz].

    Returns
    -------
    alpha, beta : floats
        The coefficient and exponent of the power law.

    """
    coeff_rkdp_dict = _coeff_rkdp_table()

    freq_band = get_freq_band(freq)
    if (freq_band is not None) and (freq_band in coeff_rkdp_dict):
        return coeff_rkdp_dict[freq_band]

    if freq < 2e9:
        freq_band_aux = 'S'
    elif freq > 12e9:
        freq_band_aux = 'X'

    warn('Radar frequency out of range. '
         + 'Coefficients only applied to S, C or X band. '
         + freq_band + ' band coefficients will be used.')

    return coeff_rkdp_dict[freq_band_aux]


def _coeff_rkdp_table():
    """
    Defines the R(kdp) power law coefficients for each frequency band.

    Returns
    -------
    coeff_rkdp_dict : dict
        A dictionary with the coefficients at each band.

    """
    coeff_rkdp_dict = dict()

    # S band: Beard and Chuang coefficients
    coeff_rkdp_dict.update({'S': (50.70, 0.8500)})

    # C band: Beard and Chuang coefficients
    coeff_rkdp_dict.update({'C': (29.70, 0.8500)})

    # X band: Brandes coefficients
    coeff_rkdp_dict.update({'X': (15.81, 0.7992)})

    return coeff_rkdp_dict


def _get_coeff_ra(freq):
    """
    Get the R(A) power law coefficients for a particular frequency.

    Parameters
    ----------
    freq : float
        Radar frequency [Hz].

    Returns
    -------
    alpha, beta : floats
        The coefficient and exponent of the power law.

    """
    coeff_ra_dict = _coeff_ra_table()

    freq_band = get_freq_band(freq)
    if (freq_band is not None) and (freq_band in coeff_ra_dict):
        return coeff_ra_dict[freq_band]

    if freq < 2e9:
        freq_band_aux = 'S'
    elif freq > 12e9:
        freq_band_aux = 'X'

    warn('Radar frequency out of range. '
         + 'Coefficients only applied to S, C or X band. '
         + freq_band + ' band coefficients will be used.')

    return coeff_ra_dict[freq_band_aux]


def _coeff_ra_table():
    """
    Defines the R(A) power law coefficients for each frequency band.

    Returns
    -------
    coeff_ra_dict : dict
        A dictionary with the coefficients at each band.

    """
    coeff_ra_dict = dict()

    # S band: at 10 deg C according to tables from Ryzhkov et al. 2014
    coeff_ra_dict.update({'S': (3100., 1.03)})

    # C band: at 10 deg C according to tables from Diederich et al. 2015
    coeff_ra_dict.update({'C': (250., 0.91)})

    # X band: at 10 deg C according to tables from Diederich et al. 2015
    coeff_ra_dict.update({'X': (45.5, 0.83)})

    return coeff_ra_dict



"""
Utilities for reading ARM KAZR spectra files.

"""

import datetime
import re

import numpy as np
try:
    import xarray as xr
    _XARRAY_AVAILABLE = True
except ImportError:
    _XARRAY_AVAILABLE = False

from ..config import FileMetadata
from ..core.radar_spectra import RadarSpectra
from ..exceptions import MissingOptionalDependency
from ..io.common import _test_arguments


KAZR_SPECTRA_FIELD_NAMES = {'spectra': 'spectra'}

def read_kazr_spectra(filename, field_names=None, additional_metadata=None,
                      file_field_names=False, exclude_fields=None,
                      include_fields=None, **kwargs):
    """
    Read a ARM KAZR spectra netCDF file.

    Parameters
    ----------
    filename : str
        Name of KAZR spectra netCDF file to read data from.
    field_names : dict, optional
        Dictionary mapping field names in the file names to radar_spectra field names.
        Unlike other read functions, fields not in this dictionary or having a
        value of None are still included in the radar_spectra.fields dictionary, to
        exclude them use the `exclude_fields` parameter. Fields which are
        mapped by this dictionary will be renamed from key to value.
    additional_metadata : dict of dicts, optional
        This parameter is not used, it is included for uniformity.
    file_field_names : bool, optional
        True to force the use of the field names from the file in which
        case the `field_names` parameter is ignored. False will use to
        `field_names` parameter to rename fields.
    exclude_fields : list or None, optional
        List of fields to exclude from the radar spectra object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields specified by include_fields.
    include_fields : list or None, optional
        List of fields to include from the radar spectra object. This is applied
        after the `file_field_names` and `field_names` parameters. Set
        to None to include all fields not specified by exclude_fields.
    delay_field_loading : bool
        True to delay loading of field data from the file until the 'data'
        key in a particular field dictionary is accessed. In this case
        the field attribute of the returned RadarSpectra object will contain
        LazyLoadDict objects not dict objects. Delayed field loading will not
        provide any speedup in file where the number of gates vary between
        rays (ngates_vary=True) and is not recommended.

    Returns
    -------
    radar_spectra : RadarSpectra
        RadarSpectra object.

    """
    if not _XARRAY_AVAILABLE:
        raise MissingOptionalDependency(
            "Xarray is required to read KAZR spectra files "
            "but is not installed!")
    # test for non empty kwargs
    _test_arguments(kwargs)

     # create metadata retrieval object
    if field_names is None:
        field_names = KAZR_SPECTRA_FIELD_NAMES
    filemetadata = FileMetadata('kazr_spectra', field_names,
                                additional_metadata,
                                file_field_names, exclude_fields)
    # read the data
    xrobj = xr.open_dataset(filename)
    # retrieve the correct time format
    # time format changed in later KAZR files
    try:
        times = [(datetime.timedelta(seconds=sec) +
                  datetime.datetime.utcfromtimestamp(
                      xrobj.base_time.values.astype("O")/1e9))
                 for sec in xrobj.time_offset.values]
    except TypeError:
        ts = [(i - np.datetime64('1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
              for i in xrobj.time_offset.values]
        bs = (
            xrobj.base_time.values - np.datetime64(
                '1970-01-01T00:00:00Z')) / np.timedelta64(1, 's')
        times = [(datetime.timedelta(seconds=sec) +
                  datetime.datetime.utcfromtimestamp(bs.astype("O")/1e9))
                 for sec in ts]
    times = np.array(times)
    time_dict = filemetadata('time')
    base_time = ('seconds since ' + datetime.datetime.strftime(
        times[0], '%Y-%m-%DT%H:%M:%SZ'))
    time_array = [(x-times[0]).seconds for x in times]
    time_dict['units'] = base_time
    time = xr.DataArray(np.array(time_array), attrs=time_dict, dims='time')

    rng_array = np.squeeze(xrobj.range.values)
    rng_dict = filemetadata('range')
    _range = xr.DataArray(rng_array, attrs=rng_dict, dims='range')

    #metadata = filemetadata('metadata')
    # Retrieve the ARM site info
    if 'nsa' in filename:
        xrobj.attrs['site_id'] = 'nsa'
    elif 'ena' in filename:
        xrobj.attrs['site_id'] = 'ena'
    else:
        xrobj.attrs['site_id'] = 'sgp'

    lat_dict = filemetadata('latitude')
    lon_dict = filemetadata('longitude')
    alt_dict = filemetadata('altitude')
    lat, lon, alt = get_kazr_location(xrobj.attrs['site_id'])
    latitude = xr.DataArray(np.array(lat, dtype='float32'), attrs=lat_dict)
    longitude = xr.DataArray(np.array(lon, dtype='float32'), attrs=lon_dict)
    altitude = xr.DataArray(np.array(alt, dtype='float32'), attrs=alt_dict)

    elevation_dict = filemetadata('elevation')
    elevation = xr.DataArray(np.array(90.0*np.ones(len(time_array)),
                                      dtype='float32'),
                             attrs=elevation_dict, dims='time')

    azimuth_dict = filemetadata('azimuth')
    azimuth = xr.DataArray(np.array(360.0*np.ones(len(time_array)),
                                    dtype='float32'),
                           attrs=azimuth_dict, dims='time')

    fix_agl_dict = filemetadata('fixed_angle')
    fixed_angle = xr.DataArray(np.array(90.0, dtype='float32'),
                               attrs=fix_agl_dict)

    sweep_number_dict = filemetadata('sweep_number')
    sweep_number = xr.DataArray(np.array([1], dtype='int32'),
                                attrs=sweep_number_dict)

    sweep_mode_dict = filemetadata('sweep_mode')
    sweep_mode = xr.DataArray(np.array('spectra'),
                              attrs=sweep_mode_dict)

    sweep_start_ray_index_dict = filemetadata('sweep_start_ray_index')
    sweep_start_ray_index = xr.DataArray(np.array([0], dtype='int32'),
                                         attrs=sweep_start_ray_index_dict)
    sweep_end_ray_index_dict = filemetadata('sweep_end_ray_index')
    sweep_end_ray_index = xr.DataArray(np.array([len(times)-1], dtype='int32'),
                                       attrs=sweep_end_ray_index_dict)
    spectra_array = [_get_spectra(xrobj, x) for x in xrobj.time.values]
    spectras = np.stack(spectra_array)

    c = 299792458

    npulses_max = xrobj.speclength.values
    wavelength = c/float(re.findall(r"[-+]?\d*\.\d+|\d+",
                                    xrobj.radar_operating_frequency)[0])*1e-9
    scan_type = 'vpt'

    metadata = xrobj.attrs
    metadata['wavelength'] = wavelength
    velocity_dict = xrobj.velocity_bins.attrs
    bins = xrobj.velocity_bins.values
    velocity_bins = xr.DataArray(bins, attrs=velocity_dict,
                                 dims='npulses_max')
    xrobj.close()

    return RadarSpectra(time, _range, spectras, metadata, scan_type,
                        latitude, longitude, altitude,
                        sweep_number, sweep_mode, fixed_angle,
                        sweep_start_ray_index, sweep_end_ray_index,
                        azimuth, elevation, npulses_max, velocity_bins)


def _get_spectra(xrobj, time):
    """ Retrieves the spectra values using time and locator mask. """
    the_spectra_locs = xrobj.locator_mask.sel(time=time).values
    the_spectra_loc = np.zeros((len(the_spectra_locs),
                                len(xrobj.speclength.values)))
    the_spectra_locs[np.isnan(the_spectra_locs)] = -9999.0
    for i, locs in enumerate(the_spectra_locs):
        if locs != -9999.0:
            the_spectra_loc[i, :] = xrobj.spectra.values[int(locs), :]
        else:
            the_spectra_loc[i, :] = np.nan*np.ones(len(
                xrobj.speclength.values))
    return the_spectra_loc


def get_kazr_location(arm_site):
    """
    Return the latitude, longitude and altitude of a ARM KAZR Radar.

    Parameters
    ----------
    arm_site : str
        Three letter ARM site acroynm.

    Returns
    -------
    lat, lon, alt : float
        Latitude (in degrees), longitude (in degrees), and altitude
        (in meters above mean sea level) of the KAZR radar.

    """
    loc = ARM_SITES[arm_site.upper()]
    return loc['lat'], loc['lon'], loc['alt']

ARM_SITES = {
    "SGP": {'lat': 36.605, 'lon': -97.485, 'alt': 318},
    "ENA": {'lat': 39.0916, 'lon': -28.0257, 'alt': 30},
    "NSA": {'lat': 71.323, 'lon': -156.609, 'alt': 8}, }
